import math

from keras import ops

from keras_hub.src.layers.modeling.rotary_embedding import RotaryEmbedding


class LlamaRotaryEmbedding(RotaryEmbedding):
    """Rotary positional encoding layer.

    This layer encodes absolute positional information with a rotation
    matrix. It calculates the rotary encoding with a mix of sine and
    cosine functions with geometrically increasing wavelengths.
    Defined and formulated in
    [RoFormer: Enhanced Transformer with Rotary Position Embedding](https://arxiv.org/abs/2104.09864v4).
    The input must be a tensor with shape a sequence dimension and a feature
    dimension. Typically, this will either an input with shape
    `(batch_size, sequence_length, feature_length)` or
    `(batch_size, sequence_length, num_heads, feature_length)`.
    This layer will return a new tensor with the rotary embedding applied to
    the input tensor.
    It is extended from `RotaryEmbedding` layer in `keras_hub.layers`.
    It has additional smoothening and interpolation for some frequency ranges.

    Args:
        max_wavelength: int. The maximum angular wavelength of the sine/cosine
            curves. Defaults to `10000`.
        position_scaling_factor: float. The scaling factor used to scale
            positions of the tokens. Defaults to `1.0`.
        frequency_adjustment_factor: float. The scaling factor used to scale the
            inverse frequencies. Defaults to `None`.
        low_freq_factor: float. The low frequency scaling factor.
            Defaults to `None`.
        high_freq_factor: float. The high frequency scaling factor.
            Defaults to `None`.
        pretraining_sequence_length: int. Used for Llama3.1+, the original
            context length at time of pretraining. Defaults to `None`.
        sequence_axis: int. Sequence axis in the input tensor.
        feature_axis: int. Feature axis in the input tensor.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `trainable`, `dtype` etc.

    Call arguments:
        inputs: The tensor inputs to apply the embedding to. This can have
            any shape, but must contain both a sequence and feature axis. The
            rotary embedding will be applied to `inputs` and returned.
        start_index: An integer or integer tensor. The starting position to
            compute the rotary embedding from. This is useful during cached
            decoding, where each position is predicted separately in a loop.

    Examples:

    ```python
    batch_size = 16
    feature_length = 18
    sequence_length = 256
    num_heads = 8

    # No multi-head dimension.
    tensor = np.ones((batch_size, sequence_length, feature_length))
    rot_emb_layer = RotaryEmbedding()
    tensor_rot = rot_emb_layer(tensor)

    # With multi-head dimension.
    tensor = np.ones((batch_size, sequence_length, num_heads, feature_length))
    tensor_rot = rot_emb_layer(tensor)
    ```

    References:
     - [RoFormer: Enhanced Transformer with Rotary Position Embedding](https://arxiv.org/abs/2104.09864v4)
    """

    def __init__(
        self,
        max_wavelength=10000,
        position_scaling_factor=1.0,
        sequence_axis=1,
        feature_axis=-1,
        frequency_adjustment_factor=None,
        low_freq_factor=None,
        high_freq_factor=None,
        pretraining_sequence_length=None,
        **kwargs,
    ):
        super().__init__(
            max_wavelength=max_wavelength,
            scaling_factor=position_scaling_factor,
            sequence_axis=sequence_axis,
            feature_axis=feature_axis,
            **kwargs,
        )
        self.max_wavelength = max_wavelength
        self.sequence_axis = sequence_axis
        self.feature_axis = feature_axis
        self.position_scaling_factor = position_scaling_factor
        self.frequency_adjustment_factor = frequency_adjustment_factor
        self.low_freq_factor = low_freq_factor
        self.high_freq_factor = high_freq_factor
        self.pretraining_sequence_length = pretraining_sequence_length

        grouped_args = [
            low_freq_factor,
            high_freq_factor,
            frequency_adjustment_factor,
            pretraining_sequence_length,
        ]
        args_none = [x is None for x in grouped_args]
        if any(args_none) and not all(args_none):
            raise ValueError(
                "Either all of `low_freq_factor`,`high_freq_factor`, "
                "`frequency_adjustment_factor` and "
                "`pretraining_sequence_length` should be set, or all of should"
                " be set `None`."
            )
        self.built = True

    def _get_inverse_freq(self, rotary_dim):
        freq_range = ops.divide(
            ops.arange(0, rotary_dim, 2, dtype="float32"),
            ops.cast(rotary_dim, "float32"),
        )
        inverse_freq = 1.0 / (self.max_wavelength**freq_range)

        # From llama3.1+ we have additional smoothening and interpolation.
        # low_freq_factor, high_freq_factor, pretraining_sequence_length,
        # frequency_adjustment_factor are all set at once so it is fine.
        if self.low_freq_factor is not None:
            low_freq_wavelen = (
                self.pretraining_sequence_length / self.low_freq_factor
            )
            high_freq_wavelen = (
                self.pretraining_sequence_length / self.high_freq_factor
            )
            wavelen = 2 * math.pi / inverse_freq

            # wavelen < high_freq_wavelen: do nothing
            # wavelen > low_freq_wavelen: divide by factor
            inverse_freq = ops.where(
                ops.greater(wavelen, low_freq_wavelen),
                (inverse_freq / self.frequency_adjustment_factor),
                inverse_freq,
            )

            # otherwise: interpolate between the two, using a smooth factor
            smooth_factor = (
                (self.pretraining_sequence_length / wavelen)
                - self.low_freq_factor
            ) / (self.high_freq_factor - self.low_freq_factor)
            smoothed_inv_freq = (1 - smooth_factor) * (
                inverse_freq / self.frequency_adjustment_factor
            ) + (smooth_factor * inverse_freq)
            is_medium_freq = ops.logical_and(
                ops.greater_equal(wavelen, high_freq_wavelen),
                ops.less_equal(wavelen, low_freq_wavelen),
            )

            inverse_freq = ops.where(
                is_medium_freq, smoothed_inv_freq, inverse_freq
            )

        return inverse_freq

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "max_wavelength": self.max_wavelength,
                "sequence_axis": self.sequence_axis,
                "feature_axis": self.feature_axis,
                "position_scaling_factor": self.position_scaling_factor,
                "frequency_adjustment_factor": self.frequency_adjustment_factor,
                "low_freq_factor": self.low_freq_factor,
                "high_freq_factor": self.high_freq_factor,
                "original_max_embeddings": self.pretraining_sequence_length,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        return input_shape



"""Convert timm models to KerasHub."""

from keras_hub.src.models.image_classifier import ImageClassifier
from keras_hub.src.utils.preset_utils import PresetLoader
from keras_hub.src.utils.preset_utils import jax_memory_cleanup
from keras_hub.src.utils.timm import convert_cspnet
from keras_hub.src.utils.timm import convert_densenet
from keras_hub.src.utils.timm import convert_efficientnet
from keras_hub.src.utils.timm import convert_mobilenet
from keras_hub.src.utils.timm import convert_resnet
from keras_hub.src.utils.timm import convert_vgg
from keras_hub.src.utils.transformers.safetensor_utils import SafetensorLoader


class TimmPresetLoader(PresetLoader):
    def __init__(self, preset, config):
        super().__init__(preset, config)
        architecture = self.config["architecture"]
        if architecture.startswith("resnet"):
            self.converter = convert_resnet
        elif architecture.startswith(("csp", "dark")):
            self.converter = convert_cspnet
        elif architecture.startswith("densenet"):
            self.converter = convert_densenet
        elif architecture.startswith("mobilenet"):
            self.converter = convert_mobilenet
        elif architecture.startswith("vgg"):
            self.converter = convert_vgg
        elif architecture.startswith("efficientnet"):
            self.converter = convert_efficientnet
        else:
            raise ValueError(
                "KerasHub has no converter for timm models "
                f"with architecture `'{architecture}'`."
            )

    def check_backbone_class(self):
        return self.converter.backbone_cls

    def load_backbone(self, cls, load_weights, **kwargs):
        keras_config = self.converter.convert_backbone_config(self.config)
        backbone = cls(**{**keras_config, **kwargs})
        if load_weights:
            jax_memory_cleanup(backbone)
            # Use prefix="" to avoid using `get_prefixed_key`.
            with SafetensorLoader(self.preset, prefix="") as loader:
                self.converter.convert_weights(backbone, loader, self.config)
        return backbone

    def load_task(self, cls, load_weights, load_task_weights, **kwargs):
        if not load_task_weights or not issubclass(cls, ImageClassifier):
            return super().load_task(
                cls, load_weights, load_task_weights, **kwargs
            )
        # Support loading the classification head for classifier models.
        kwargs["num_classes"] = self.config["num_classes"]
        if (
            "num_features" in self.config
            and "mobilenet" in self.config["architecture"]
        ):
            kwargs["num_features"] = self.config["num_features"]

        task = super().load_task(cls, load_weights, load_task_weights, **kwargs)
        if load_task_weights:
            with SafetensorLoader(self.preset, prefix="") as loader:
                self.converter.convert_head(task, loader, self.config)
        return task

    def load_image_converter(self, cls, **kwargs):
        pretrained_cfg = self.config.get("pretrained_cfg", None)
        if not pretrained_cfg or "input_size" not in pretrained_cfg:
            return None
        # This assumes the same basic setup for all timm preprocessing, We may
        # need to extend this as we cover more model types.
        input_size = pretrained_cfg["input_size"]
        mean = pretrained_cfg["mean"]
        std = pretrained_cfg["std"]
        scale = [1.0 / 255.0 / s for s in std]
        offset = [-m / s for m, s in zip(mean, std)]
        interpolation = pretrained_cfg["interpolation"]
        if interpolation not in ("bilinear", "nearest", "bicubic"):
            interpolation = "bilinear"  # Unsupported interpolation type.
        return cls(
            image_size=input_size[1:],
            scale=scale,
            offset=offset,
            interpolation=interpolation,
        )



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_segmenter import ImageSegmenter
from keras_hub.src.models.segformer.segformer_backbone import SegFormerBackbone
from keras_hub.src.models.segformer.segformer_image_segmenter_preprocessor import (  # noqa: E501
    SegFormerImageSegmenterPreprocessor,
)


@keras_hub_export("keras_hub.models.SegFormerImageSegmenter")
class SegFormerImageSegmenter(ImageSegmenter):
    """A Keras model implementing SegFormer for semantic segmentation.

    This class implements the segmentation head of the SegFormer architecture
    described in [SegFormer: Simple and Efficient Design for Semantic
    Segmentation with Transformers] (https://arxiv.org/abs/2105.15203) and
    [based on the TensorFlow implementation from DeepVision]
    (https://github.com/DavidLandup0/deepvision/tree/main/deepvision/models/segmentation/segformer).

    SegFormers are meant to be used with the MixTransformer (MiT) encoder
    family, and and use a very lightweight all-MLP decoder head.

    The MiT encoder uses a hierarchical transformer which outputs features at
    multiple scales, similar to that of the hierarchical outputs typically
    associated with CNNs.

    Args:
        image_encoder: `keras.Model`. The backbone network for the model that is
            used as a feature extractor for the SegFormer encoder. It is
            *intended* to be used only with the MiT backbone model
            (`keras_hub.models.MiTBackbone`) which was created specifically for
            SegFormers. Alternatively, can be a `keras_hub.models.Backbone` a
            model subclassing `keras_hub.models.FeaturePyramidBackbone`, or a
            `keras.Model` that has a `pyramid_outputs` property which is a
            dictionary with keys "P2", "P3", "P4", and "P5" and layer names as
            values.
        num_classes: int, the number of classes for the detection model,
            including the background class.
        projection_filters: int, number of filters in the
            convolution layer projecting the concatenated features into a
            segmentation map. Defaults to 256`.


    Example:

    Using presets:

    ```python
    segmenter = keras_hub.models.SegFormerImageSegmenter.from_preset(
        "segformer_b0_ade20k_512"
    )

    images = np.random.rand(1, 512, 512, 3)
    segformer(images)
    ```

    Using the SegFormer backbone:

    ```python
    encoder = keras_hub.models.MiTBackbone.from_preset(
        "mit_b0_ade20k_512"
    )
    backbone = keras_hub.models.SegFormerBackbone(
        image_encoder=encoder,
        projection_filters=256,
    )
    ```

    Using the SegFormer backbone with a custom encoder:

    ```python
    images = np.ones(shape=(1, 96, 96, 3))
    labels = np.zeros(shape=(1, 96, 96, 1))

    encoder = keras_hub.models.MiTBackbone(
        depths=[2, 2, 2, 2],
        image_shape=(96, 96, 3),
        hidden_dims=[32, 64, 160, 256],
        num_layers=4,
        blockwise_num_heads=[1, 2, 5, 8],
        blockwise_sr_ratios=[8, 4, 2, 1],
        max_drop_path_rate=0.1,
        patch_sizes=[7, 3, 3, 3],
        strides=[4, 2, 2, 2],
    )

    backbone = keras_hub.models.SegFormerBackbone(
        image_encoder=encoder,
        projection_filters=256,
    )
    segformer = keras_hub.models.SegFormerImageSegmenter(
        backbone=backbone,
        num_classes=4,
    )
    segformer(images
    ```

    Using the segmentor class with a preset backbone:

    ```python
    image_encoder = keras_hub.models.MiTBackbone.from_preset(
        "mit_b0_ade20k_512"
    )
    backbone = keras_hub.models.SegFormerBackbone(
        image_encoder=encoder,
        projection_filters=256,
    )
    segformer = keras_hub.models.SegFormerImageSegmenter(
        backbone=backbone,
        num_classes=4,
    )
    ```
    """

    backbone_cls = SegFormerBackbone
    preprocessor_cls = SegFormerImageSegmenterPreprocessor

    def __init__(
        self,
        backbone,
        num_classes,
        preprocessor=None,
        **kwargs,
    ):
        if not isinstance(backbone, keras.layers.Layer) or not isinstance(
            backbone, keras.Model
        ):
            raise ValueError(
                "Argument `backbone` must be a `keras.layers.Layer` instance "
                f" or `keras.Model`. Received instead "
                f"backbone={backbone} (of type {type(backbone)})."
            )

        # === Layers ===
        inputs = backbone.input

        self.backbone = backbone
        self.preprocessor = preprocessor
        self.dropout = keras.layers.Dropout(0.1)
        self.output_segmentation_head = keras.layers.Conv2D(
            filters=num_classes, kernel_size=1, strides=1
        )
        self.resizing = keras.layers.Resizing(
            height=inputs.shape[1],
            width=inputs.shape[2],
            interpolation="bilinear",
        )

        # === Functional Model ===
        x = self.backbone(inputs)
        x = self.dropout(x)
        x = self.output_segmentation_head(x)
        output = self.resizing(x)

        super().__init__(
            inputs=inputs,
            outputs=output,
            **kwargs,
        )

        # === Config ===
        self.num_classes = num_classes
        self.backbone = backbone

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "num_classes": self.num_classes,
                "backbone": keras.saving.serialize_keras_object(self.backbone),
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        if "image_encoder" in config and isinstance(
            config["image_encoder"], dict
        ):
            config["image_encoder"] = keras.layers.deserialize(
                config["image_encoder"]
            )
        return super().from_config(config)



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.preprocessing_layer import (
    PreprocessingLayer,
)
from keras_hub.src.utils.preset_utils import PREPROCESSOR_CONFIG_FILE
from keras_hub.src.utils.preset_utils import builtin_presets
from keras_hub.src.utils.preset_utils import find_subclass
from keras_hub.src.utils.preset_utils import get_preset_loader
from keras_hub.src.utils.preset_utils import get_preset_saver
from keras_hub.src.utils.python_utils import classproperty


@keras_hub_export("keras_hub.models.Preprocessor")
class Preprocessor(PreprocessingLayer):
    """Base class for preprocessing layers.

    A `Preprocessor` layer provides a complete preprocessing setup for a
    given task. It handles tokenization, audio/image conversion, and
    any other necessary preprocessing steps.

    This class can be subclassed similar to any `keras.layers.Layer`, by
    defining `build()`, `call()` and `get_config()` methods. All subclasses
    should set the `tokenizer` or `audio_converter` or `image_converter`
    properties during construction as needed.
    """

    backbone_cls = None
    tokenizer_cls = None
    audio_converter_cls = None
    image_converter_cls = None

    def __init__(self, *args, **kwargs):
        self.config_file = kwargs.pop("config_file", PREPROCESSOR_CONFIG_FILE)
        super().__init__(*args, **kwargs)
        self._tokenizer = None
        self._image_converter = None
        self._audio_converter = None

    def __setattr__(self, name, value):
        # Work around torch setattr for properties.
        if name in ["tokenizer", "audio_converter", "image_converter"]:
            return object.__setattr__(self, name, value)
        return super().__setattr__(name, value)

    @property
    def tokenizer(self):
        """The tokenizer used to tokenize strings."""
        return self._tokenizer

    @tokenizer.setter
    def tokenizer(self, value):
        self._tokenizer = value

    @property
    def audio_converter(self):
        """The audio converter used to preprocess audio data."""
        return self._audio_converter

    @audio_converter.setter
    def audio_converter(self, value):
        self._audio_converter = value

    @property
    def image_converter(self):
        """The image converter used to preprocess image data."""
        return self._image_converter

    @image_converter.setter
    def image_converter(self, value):
        self._image_converter = value

    @property
    def image_size(self):
        """Shortcut to get/set the image size of the image converter."""
        if self.image_converter is None:
            return None
        return self.image_converter.image_size

    @image_size.setter
    def image_size(self, value):
        if self.image_converter is None:
            raise ValueError(
                "Cannot set `image_size` on preprocessor if `image_converter` "
                " is `None`."
            )
        self.image_converter.image_size = value

    def get_config(self):
        config = super().get_config()
        if self.tokenizer:
            config["tokenizer"] = keras.layers.serialize(self.tokenizer)
        if self.audio_converter:
            config["audio_converter"] = keras.layers.serialize(
                self.audio_converter
            )
        if self.image_converter:
            config["image_converter"] = keras.layers.serialize(
                self.image_converter
            )
        config.update(
            {
                "config_file": self.config_file,
            }
        )
        return config

    @classmethod
    def from_config(cls, config):
        if "tokenizer" in config and isinstance(config["tokenizer"], dict):
            config["tokenizer"] = keras.layers.deserialize(config["tokenizer"])
        if "audio_converter" in config and isinstance(
            config["audio_converter"], dict
        ):
            config["audio_converter"] = keras.layers.deserialize(
                config["audio_converter"]
            )
        if "image_converter" in config and isinstance(
            config["image_converter"], dict
        ):
            config["image_converter"] = keras.layers.deserialize(
                config["image_converter"]
            )
        return cls(**config)

    @classproperty
    def presets(cls):
        """List built-in presets for a `Preprocessor` subclass."""
        return builtin_presets(cls)

    @classmethod
    def from_preset(
        cls,
        preset,
        config_file=PREPROCESSOR_CONFIG_FILE,
        **kwargs,
    ):
        """Instantiate a `keras_hub.models.Preprocessor` from a model preset.

        A preset is a directory of configs, weights and other file assets used
        to save and load a pre-trained model. The `preset` can be passed as
        one of:

        1. a built-in preset identifier like `'bert_base_en'`
        2. a Kaggle Models handle like `'kaggle://user/bert/keras/bert_base_en'`
        3. a Hugging Face handle like `'hf://user/bert_base_en'`
        4. a path to a local preset directory like `'./bert_base_en'`

        For any `Preprocessor` subclass, you can run `cls.presets.keys()` to
        list all built-in presets available on the class.

        As there are usually multiple preprocessing classes for a given model,
        this method should be called on a specific subclass like
        `keras_hub.models.BertTextClassifierPreprocessor.from_preset()`.

        Args:
            preset: string. A built-in preset identifier, a Kaggle Models
                handle, a Hugging Face handle, or a path to a local directory.

        Examples:
        ```python
        # Load a preprocessor for Gemma generation.
        preprocessor = keras_hub.models.CausalLMPreprocessor.from_preset(
            "gemma_2b_en",
        )

        # Load a preprocessor for Bert classification.
        preprocessor = keras_hub.models.TextClassifierPreprocessor.from_preset(
            "bert_base_en",
        )
        ```
        """
        if cls == Preprocessor:
            raise ValueError(
                "Do not call `Preprocessor.from_preset()` directly. Instead "
                "choose a particular task preprocessing class, e.g. "
                "`keras_hub.models.TextClassifierPreprocessor.from_preset()`."
            )

        loader = get_preset_loader(preset)
        backbone_cls = loader.check_backbone_class()
        # Detect the correct subclass if we need to.
        if cls.backbone_cls != backbone_cls:
            cls = find_subclass(preset, cls, backbone_cls)
        return loader.load_preprocessor(cls, config_file, **kwargs)

    @classmethod
    def _add_missing_kwargs(cls, loader, kwargs):
        """Fill in required kwargs when loading from preset.

        This is a private method hit when loading a preprocessing layer that
        was not directly saved in the preset. This method should fill in
        all required kwargs required to call the class constructor. For almost,
        all preprocessors, the only required args are `tokenizer`,
        `image_converter`, and `audio_converter`, but this can be overridden,
        e.g. for a preprocessor with multiple tokenizers for different
        encoders.
        """
        if "tokenizer" not in kwargs and cls.tokenizer_cls:
            kwargs["tokenizer"] = loader.load_tokenizer(cls.tokenizer_cls)
        if "audio_converter" not in kwargs and cls.audio_converter_cls:
            kwargs["audio_converter"] = loader.load_audio_converter(
                cls.audio_converter_cls
            )
        if "image_converter" not in kwargs and cls.image_converter_cls:
            kwargs["image_converter"] = loader.load_image_converter(
                cls.image_converter_cls
            )
        return kwargs

    def load_preset_assets(self, preset):
        """Load all static assets needed by the preprocessing layer.

        Args:
            preset_dir: The path to the local model preset directory.
        """
        for layer in self._flatten_layers(include_self=False):
            if hasattr(layer, "load_preset_assets"):
                layer.load_preset_assets(preset)

    def save_to_preset(self, preset_dir):
        """Save preprocessor to a preset directory.

        Args:
            preset_dir: The path to the local model preset directory.
        """
        saver = get_preset_saver(preset_dir)
        saver.save_preprocessor(self)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.object_detector_preprocessor import (
    ObjectDetectorPreprocessor,
)
from keras_hub.src.models.retinanet.retinanet_backbone import RetinaNetBackbone
from keras_hub.src.models.retinanet.retinanet_image_converter import (
    RetinaNetImageConverter,
)


@keras_hub_export("keras_hub.models.RetinaNetObjectDetectorPreprocessor")
class RetinaNetObjectDetectorPreprocessor(ObjectDetectorPreprocessor):
    backbone_cls = RetinaNetBackbone
    image_converter_cls = RetinaNetImageConverter



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)
from keras_hub.src.models.mit.mit_backbone import MiTBackbone
from keras_hub.src.models.mit.mit_image_converter import MiTImageConverter


@keras_hub_export("keras_hub.models.MiTImageClassifierPreprocessor")
class MiTImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = MiTBackbone
    image_converter_cls = MiTImageConverter



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.deberta_v3.deberta_v3_backbone import (
    DebertaV3Backbone,
)
from keras_hub.src.tokenizers.sentence_piece_tokenizer import (
    SentencePieceTokenizer,
)

try:
    import tensorflow as tf
except ImportError:
    tf = None


@keras_hub_export(
    [
        "keras_hub.tokenizers.DebertaV3Tokenizer",
        "keras_hub.models.DebertaV3Tokenizer",
    ]
)
class DebertaV3Tokenizer(SentencePieceTokenizer):
    """DeBERTa tokenizer layer based on SentencePiece.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_hub.tokenizers.SentencePieceTokenizer`. Unlike the
    underlying tokenizer, it will check for all special tokens needed by
    DeBERTa models and provides a `from_preset()` method to automatically
    download a matching vocabulary for a DeBERTa preset.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.

    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Note: The mask token (`"[MASK]"`) is handled differently in this tokenizer.
    If the token is not present in the provided SentencePiece vocabulary, the
    token will be appended to the vocabulary. For example, if the vocabulary
    size is 100, the mask token will be assigned the ID 100.

    Args:
        proto: Either a `string` path to a SentencePiece proto file, or a
            `bytes` object with a serialized SentencePiece proto. See the
            [SentencePiece repository](https://github.com/google/sentencepiece)
            for more details on the format.

    Examples:

    ```python
    # Unbatched input.
    tokenizer = keras_hub.models.DebertaV3Tokenizer.from_preset(
        "deberta_v3_base_en",
    )
    tokenizer("The quick brown fox jumped.")

    # Batched inputs.
    tokenizer(["the quick brown fox", "the earth is round"])

    # Detokenization.
    tokenizer.detokenize(tokenizer("The quick brown fox jumped."))

    # Custom vocabulary.
    bytes_io = io.BytesIO()
    ds = tf.data.Dataset.from_tensor_slices(["The quick brown fox jumped."])
    sentencepiece.SentencePieceTrainer.train(
        sentence_iterator=ds.as_numpy_iterator(),
        model_writer=bytes_io,
        vocab_size=9,
        model_type="WORD",
        pad_id=0,
        bos_id=1,
        eos_id=2,
        unk_id=3,
        pad_piece="[PAD]",
        bos_piece="[CLS]",
        eos_piece="[SEP]",
        unk_piece="[UNK]",
    )
    tokenizer = keras_hub.models.DebertaV3Tokenizer(
        proto=bytes_io.getvalue(),
    )
    tokenizer("The quick brown fox jumped.")
    ```
    """

    backbone_cls = DebertaV3Backbone

    def __init__(self, proto, **kwargs):
        self._add_special_token("[CLS]", "cls_token")
        self._add_special_token("[SEP]", "sep_token")
        self._add_special_token("[PAD]", "pad_token")
        # Also add `tokenizer.start_token` and `tokenizer.end_token` for
        # compatibility with other tokenizers.
        self._add_special_token("[CLS]", "start_token")
        self._add_special_token("[SEP]", "end_token")
        # Handle mask separately as it's not always in the vocab.
        self.mask_token = "[MASK]"
        self.mask_token_id = None
        super().__init__(proto=proto, **kwargs)

    @property
    def special_tokens(self):
        return super().special_tokens + [self.mask_token]

    @property
    def special_token_ids(self):
        return super().special_token_ids + [self.mask_token_id]

    def set_proto(self, proto):
        super().set_proto(proto)
        if proto is not None:
            if self.mask_token in super().get_vocabulary():
                self.mask_token_id = super().token_to_id(self.mask_token)
            else:
                self.mask_token_id = super().vocabulary_size()
        else:
            self.mask_token_id = None

    def vocabulary_size(self):
        sentence_piece_size = super().vocabulary_size()
        if sentence_piece_size == self.mask_token_id:
            return sentence_piece_size + 1
        return sentence_piece_size

    def get_vocabulary(self):
        sentence_piece_vocabulary = super().get_vocabulary()
        if self.mask_token_id is None:
            return sentence_piece_vocabulary
        if self.mask_token_id < super().vocabulary_size():
            return sentence_piece_vocabulary
        return sentence_piece_vocabulary + ["[MASK]"]

    def id_to_token(self, id):
        if id == self.mask_token_id:
            return "[MASK]"
        return super().id_to_token(id)

    def token_to_id(self, token):
        if token == "[MASK]":
            return self.mask_token_id
        return super().token_to_id(token)

    def detokenize(self, ids):
        ids = tf.ragged.boolean_mask(ids, tf.not_equal(ids, self.mask_token_id))
        return super().detokenize(ids)



import numpy as np

from keras_hub.src.models.bert.bert_backbone import BertBackbone
from keras_hub.src.utils.preset_utils import HF_TOKENIZER_CONFIG_FILE
from keras_hub.src.utils.preset_utils import get_file
from keras_hub.src.utils.preset_utils import load_json

backbone_cls = BertBackbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "num_layers": transformers_config["num_hidden_layers"],
        "num_heads": transformers_config["num_attention_heads"],
        "hidden_dim": transformers_config["hidden_size"],
        "intermediate_dim": transformers_config["intermediate_size"],
    }


def convert_weights(backbone, loader, transformers_config):
    # Embedding layer
    loader.port_weight(
        keras_variable=backbone.get_layer("token_embedding").embeddings,
        hf_weight_key="bert.embeddings.word_embeddings.weight",
    )
    loader.port_weight(
        keras_variable=backbone.get_layer(
            "position_embedding"
        ).position_embeddings,
        hf_weight_key="bert.embeddings.position_embeddings.weight",
    )
    loader.port_weight(
        keras_variable=backbone.get_layer("segment_embedding").embeddings,
        hf_weight_key="bert.embeddings.token_type_embeddings.weight",
    )
    loader.port_weight(
        keras_variable=backbone.get_layer("embeddings_layer_norm").beta,
        hf_weight_key="bert.embeddings.LayerNorm.beta",
    )
    loader.port_weight(
        keras_variable=backbone.get_layer("embeddings_layer_norm").gamma,
        hf_weight_key="bert.embeddings.LayerNorm.gamma",
    )

    def transpose_and_reshape(x, shape):
        return np.reshape(np.transpose(x), shape)

    # Attention blocks
    for i in range(backbone.num_layers):
        block = backbone.get_layer(f"transformer_layer_{i}")
        attn = block._self_attention_layer
        hf_prefix = "bert.encoder.layer."
        # Attention layers
        loader.port_weight(
            keras_variable=attn.query_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.query.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.query_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.query.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        loader.port_weight(
            keras_variable=attn.key_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.key.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.key_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.key.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        loader.port_weight(
            keras_variable=attn.value_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.value.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.value_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.self.value.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        loader.port_weight(
            keras_variable=attn.output_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.attention.output.dense.weight",
            hook_fn=transpose_and_reshape,
        )
        loader.port_weight(
            keras_variable=attn.output_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.attention.output.dense.bias",
            hook_fn=lambda hf_tensor, shape: np.reshape(hf_tensor, shape),
        )
        # Attention layer norm.
        loader.port_weight(
            keras_variable=block._self_attention_layer_norm.beta,
            hf_weight_key=f"{hf_prefix}{i}.attention.output.LayerNorm.beta",
        )
        loader.port_weight(
            keras_variable=block._self_attention_layer_norm.gamma,
            hf_weight_key=f"{hf_prefix}{i}.attention.output.LayerNorm.gamma",
        )
        # MLP layers
        loader.port_weight(
            keras_variable=block._feedforward_intermediate_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.intermediate.dense.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=block._feedforward_intermediate_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.intermediate.dense.bias",
        )
        loader.port_weight(
            keras_variable=block._feedforward_output_dense.kernel,
            hf_weight_key=f"{hf_prefix}{i}.output.dense.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=block._feedforward_output_dense.bias,
            hf_weight_key=f"{hf_prefix}{i}.output.dense.bias",
        )
        # Output layer norm.
        loader.port_weight(
            keras_variable=block._feedforward_layer_norm.beta,
            hf_weight_key=f"{hf_prefix}{i}.output.LayerNorm.beta",
        )
        loader.port_weight(
            keras_variable=block._feedforward_layer_norm.gamma,
            hf_weight_key=f"{hf_prefix}{i}.output.LayerNorm.gamma",
        )

    loader.port_weight(
        keras_variable=backbone.get_layer("pooled_dense").kernel,
        hf_weight_key="bert.pooler.dense.weight",
        hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
    )
    loader.port_weight(
        keras_variable=backbone.get_layer("pooled_dense").bias,
        hf_weight_key="bert.pooler.dense.bias",
    )


def convert_tokenizer(cls, preset, **kwargs):
    transformers_config = load_json(preset, HF_TOKENIZER_CONFIG_FILE)
    return cls(
        get_file(preset, "vocab.txt"),
        lowercase=transformers_config["do_lower_case"],
        **kwargs,
    )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.layers.preprocessing.image_converter import ImageConverter
from keras_hub.src.models.segformer.segformer_backbone import SegFormerBackbone


@keras_hub_export("keras_hub.layers.SegFormerImageConverter")
class SegFormerImageConverter(ImageConverter):
    backbone_cls = SegFormerBackbone



"""CSPNet preset configurations."""

backbone_presets = {
    "csp_darknet_53_ra_imagenet": {
        "metadata": {
            "description": (
                "A CSP-DarkNet (Cross-Stage-Partial) image classification model"
                " pre-trained on the Randomly Augmented ImageNet 1k dataset at "
                "a 256x256 resolution."
            ),
            "params": 27642184,
            "path": "cspnet",
        },
        "kaggle_handle": "kaggle://keras/cspdarknet/keras/csp_darknet_53_ra_imagenet/2",
    },
    "csp_resnext_50_ra_imagenet": {
        "metadata": {
            "description": (
                "A CSP-ResNeXt (Cross-Stage-Partial) image classification model"
                " pre-trained on the Randomly Augmented ImageNet 1k dataset at "
                "a 256x256 resolution."
            ),
            "params": 20569896,
            "path": "cspnet",
        },
        "kaggle_handle": "kaggle://keras/cspdarknet/keras/csp_resnext_50_ra_imagenet/1",
    },
    "csp_resnet_50_ra_imagenet": {
        "metadata": {
            "description": (
                "A CSP-ResNet (Cross-Stage-Partial) image classification model"
                " pre-trained on the Randomly Augmented ImageNet 1k dataset at "
                "a 256x256 resolution."
            ),
            "params": 21616168,
            "path": "cspnet",
        },
        "kaggle_handle": "kaggle://keras/cspdarknet/keras/csp_resnet_50_ra_imagenet/1",
    },
    "darknet_53_imagenet": {
        "metadata": {
            "description": (
                "A DarkNet image classification model pre-trained on the"
                "ImageNet 1k dataset at a 256x256 resolution."
            ),
            "params": 41609928,
            "path": "cspnet",
        },
        "kaggle_handle": "kaggle://keras/cspdarknet/keras/darknet_53_imagenet/1",
    },
}



import math

from keras import layers
from keras import ops

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone


class CLIPVisionPooler(layers.Layer):
    """The vision pooler layer of CLIP.

    `CLIPVisionPooler` will extracts the first token (index `0`) from the
    sequence of the vision embeddings as the pooled outputs.

    Call arguments:
        vision_embeddings: A tensor of shape
            `(batch_size, sequence_length, hidden_dim)`.
    """

    def call(self, vision_embeddings):
        return vision_embeddings[:, 0, :]

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[-1])


class CLIPTextPooler(layers.Layer):
    """The text pooler layer of CLIP.

    `CLIPTextPooler` extracts the text embeddings at the positions of EOS tokens
    as the pooled outputs.

    Call arguments:
        text_embeddings: A tensor of shape
            `(batch_size, sequence_length, hidden_dim)`.
        token_ids: A tensor of shape `(batch_size, max_tokens)`, used to
            identify the positions of EOS tokens.
    """

    def call(self, text_embeddings, token_ids):
        # `keepdims` is not supported in `keras<=3.1`.
        eos_index = ops.argmax(token_ids, axis=-1)
        eos_index = ops.expand_dims(eos_index, axis=-1)
        eos_index = ops.expand_dims(eos_index, axis=-1)
        pooled_outputs = ops.take_along_axis(text_embeddings, eos_index, axis=1)
        return ops.squeeze(pooled_outputs, axis=1)

    def compute_output_shape(self, input_shape):
        return (input_shape[0], input_shape[-1])


class CLIPHead(layers.Layer):
    """The head layer of CLIP.

    `CLIPHead` takes `vision_embedding` and `text_embedding` as inputs to
    compute the corresponding logits. Both embeddings are L2 normalized and used
    to compute pairwise cosine similarity. The resulting logits are then scaled
    by a learnable `logit_scale` parameter.

    Call arguments:
        vision_embedding: A tensor of shape `(batch_size, hidden_dim)`.
        text_embedding: A tensor of shape `(batch_size, hidden_dim)`.
    """

    def build(self, input_shape):
        self.logit_scale = self.add_weight(
            shape=(),
            initializer=lambda *a, **kw: math.log(1 / 0.07),
            trainable=True,
            dtype=self.variable_dtype,
            name="logit_scale",
        )

    def call(self, vision_embedding, text_embedding):
        normalized_vision_embedding = ops.sqrt(
            ops.sum(ops.power(vision_embedding, 2), axis=-1, keepdims=True)
        )
        normalized_text_embedding = ops.sqrt(
            ops.sum(ops.power(text_embedding, 2), axis=-1, keepdims=True)
        )
        vision_embedding = vision_embedding / normalized_vision_embedding
        text_embedding = text_embedding / normalized_text_embedding
        logit_scale = ops.exp(self.logit_scale)
        text_logits = (
            ops.matmul(
                text_embedding,
                ops.transpose(vision_embedding),
            )
            * logit_scale
        )
        vision_logits = ops.transpose(text_logits)
        return vision_logits, text_logits

    def compute_output_shape(
        self, vision_embedding_shape, text_embedding_shape
    ):
        vision_logits_shape = (
            vision_embedding_shape[0],
            text_embedding_shape[0],
        )
        text_logits_shape = (
            text_embedding_shape[0],
            vision_embedding_shape[0],
        )
        return vision_logits_shape, text_logits_shape


@keras_hub_export("keras_hub.models.CLIPBackbone")
class CLIPBackbone(Backbone):
    """CLIP core network with hyperparameters.

    This backbone implements the base architecture for Contrastive
    Language-Image Pretraining (CLIP) model. It includes a vision and text
    encoders and the corresponding projection layers. This backbone will output
    the final logit scores corresponding to each image and token input. These
    values are cosine similarities between the corresponding image and text
    features.

    The default constructor gives a fully customizable, randomly initialized
    CLIP model with any number of layers, heads, and embedding dimensions. To
    load preset architectures and weights, use the `from_preset` constructor.

    Args:
        vision_encoder: The CLIP vision encoder for encoding the input images.
        text_encoder: The CLIP text encoder for encoding the input tokens.
        projection_dim: int. The size of the projection layer.
        dtype: string or `keras.mixed_precision.DTypePolicy`. The dtype to use
            for the models computations and weights. Note that some
            computations, such as softmax and layer normalization will always
            be done a float32 precision regardless of dtype.

    Example:
    ```python
    input_data = {
        "images": np.ones(shape=(1, 224, 224, 3), dtype="float32"),
        "token_ids": np.ones(shape=(1, 12), dtype="int32"),
    }

    # Pretrained CLIP model.
    model = keras_hub.models.CLIPBackbone.from_preset("clip_vit_base_patch32")
    model(input_data)

    # Randomly initialized CLIP model with custom config.
    vision_encoder = keras_hub.models.CLIPVisionEncoder(
        patch_size=32,
        hidden_dim=768,
        num_layers=8,
        num_heads=8,
        intermediate_dim=2048,
        image_shape=(384, 384, 3),
    )
    text_encoder = keras_hub.models.CLIPTextEncoder(
        vocabulary_size=49408,
        embedding_dim=768,
        hidden_dim=768,
        num_layers=8,
        num_heads=8,
        intermediate_dim=2048,
    )
    model = keras_hub.models.CLIPBackbone(
        vision_encoder=vision_encoder,
        text_encoder=text_encoder,
        projection_dim=256,
    )
    model(input_data)
    ```
    """

    def __init__(
        self,
        vision_encoder,
        text_encoder,
        projection_dim,
        dtype=None,
        name=None,
        **kwargs,
    ):
        # === Layers ===
        self.vision_encoder = vision_encoder
        self.text_encoder = text_encoder
        self.vision_pooler = CLIPVisionPooler(dtype=dtype, name="vision_pooler")
        self.text_pooler = CLIPTextPooler(dtype=dtype, name="text_pooler")
        self.vision_projection = layers.Dense(
            projection_dim,
            use_bias=False,
            dtype=dtype,
            name="vision_projection",
        )
        self.text_projection = layers.Dense(
            projection_dim,
            use_bias=False,
            dtype=dtype,
            name="text_projection",
        )
        self.clip_head = CLIPHead(dtype=dtype, name="clip_head")

        # === Functional Model ===
        image_input = layers.Input(
            shape=self.vision_encoder.image_shape, name="images"
        )
        token_id_input = layers.Input(
            shape=(None,), dtype="int32", name="token_ids"
        )
        vision_embeddings = self.get_vision_embeddings(image_input)
        text_embeddings = self.get_text_embeddings(token_id_input)
        vision_logits, text_logits = self.clip_head(
            vision_embeddings, text_embeddings
        )

        super().__init__(
            inputs={
                "images": image_input,
                "token_ids": token_id_input,
            },
            outputs={
                "vision_logits": vision_logits,
                "text_logits": text_logits,
            },
            dtype=dtype,
            name=name,
            **kwargs,
        )

        # === Config ===
        self.projection_dim = projection_dim

    def get_vision_embeddings(self, images):
        """Get the embeddings from the vision encoder.

        Args:
            images: The input tensor for the vision encoder.

        Returns:
            The output embeddings obtained by applying projection layer to the
            pooled output of the vision encoder.
        """
        vision_outputs = self.vision_encoder({"images": images})
        vision_outputs = self.vision_pooler(vision_outputs)
        return self.vision_projection(vision_outputs)

    def get_text_embeddings(self, token_ids):
        """Get the embeddings from the text encoder.

        Args:
            token_ids: The input int tensor for the text encoder.

        Returns:
            The output embeddings obtained by applying projection layer to the
            pooled output of the text encoder.
        """
        text_outputs = self.text_encoder({"token_ids": token_ids})
        text_outputs = self.text_pooler(text_outputs, token_ids)
        return self.text_projection(text_outputs)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vision_encoder": layers.serialize(self.vision_encoder),
                "text_encoder": layers.serialize(self.text_encoder),
                "projection_dim": self.projection_dim,
            }
        )
        return config

    @classmethod
    def from_config(cls, config, custom_objects=None):
        config = config.copy()

        # Propagate `dtype` to submodels if needed.
        if "dtype" in config and config["dtype"] is not None:
            dtype_config = config["dtype"]
            if "dtype" not in config["vision_encoder"]["config"]:
                config["vision_encoder"]["config"]["dtype"] = dtype_config
            if "dtype" not in config["text_encoder"]["config"]:
                config["text_encoder"]["config"]["dtype"] = dtype_config

        # We expect submodels to be instantiated.
        config["vision_encoder"] = layers.deserialize(
            config["vision_encoder"], custom_objects=custom_objects
        )
        config["text_encoder"] = layers.deserialize(
            config["text_encoder"], custom_objects=custom_objects
        )
        return cls(**config)



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.causal_lm_preprocessor import CausalLMPreprocessor
from keras_hub.src.models.qwen.qwen_backbone import QwenBackbone
from keras_hub.src.models.qwen.qwen_tokenizer import QwenTokenizer


@keras_hub_export(
    [
        "keras_hub.models.QwenCausalLMPreprocessor",
        "keras_hub.models.Qwen2CausalLMPreprocessor",
    ]
)
class QwenCausalLMPreprocessor(CausalLMPreprocessor):
    backbone_cls = QwenBackbone
    tokenizer_cls = QwenTokenizer

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import math

import numpy as np

from keras_hub.src.models.efficientnet.efficientnet_backbone import (
    EfficientNetBackbone,
)

backbone_cls = EfficientNetBackbone


VARIANT_MAP = {
    "b0": {
        "stackwise_width_coefficients": [1.0] * 7,
        "stackwise_depth_coefficients": [1.0] * 7,
        "stackwise_squeeze_and_excite_ratios": [0.25] * 7,
    },
    "b1": {
        "stackwise_width_coefficients": [1.0] * 7,
        "stackwise_depth_coefficients": [1.1] * 7,
        "stackwise_squeeze_and_excite_ratios": [0.25] * 7,
    },
    "b2": {
        "stackwise_width_coefficients": [1.1] * 7,
        "stackwise_depth_coefficients": [1.2] * 7,
        "stackwise_squeeze_and_excite_ratios": [0.25] * 7,
    },
    "b3": {
        "stackwise_width_coefficients": [1.2] * 7,
        "stackwise_depth_coefficients": [1.4] * 7,
        "stackwise_squeeze_and_excite_ratios": [0.25] * 7,
    },
    "b4": {
        "stackwise_width_coefficients": [1.4] * 7,
        "stackwise_depth_coefficients": [1.8] * 7,
        "stackwise_squeeze_and_excite_ratios": [0.25] * 7,
    },
    "b5": {
        "stackwise_width_coefficients": [1.6] * 7,
        "stackwise_depth_coefficients": [2.2] * 7,
        "stackwise_squeeze_and_excite_ratios": [0.25] * 7,
    },
    "lite0": {
        "stackwise_width_coefficients": [1.0] * 7,
        "stackwise_depth_coefficients": [1.0] * 7,
        "stackwise_squeeze_and_excite_ratios": [0] * 7,
        "activation": "relu6",
    },
    "el": {
        "stackwise_width_coefficients": [1.2] * 6,
        "stackwise_depth_coefficients": [1.4] * 6,
        "stackwise_kernel_sizes": [3, 3, 3, 5, 5, 5],
        "stackwise_num_repeats": [1, 2, 4, 5, 4, 2],
        "stackwise_input_filters": [32, 24, 32, 48, 96, 144],
        "stackwise_output_filters": [24, 32, 48, 96, 144, 192],
        "stackwise_expansion_ratios": [4, 8, 8, 8, 8, 8],
        "stackwise_strides": [1, 2, 2, 2, 1, 2],
        "stackwise_squeeze_and_excite_ratios": [0] * 6,
        "stackwise_block_types": ["fused"] * 3 + ["unfused"] * 3,
        "stackwise_force_input_filters": [24, 0, 0, 0, 0, 0],
        "stackwise_nores_option": [True] + [False] * 5,
        "activation": "relu",
    },
    "em": {
        "stackwise_width_coefficients": [1.0] * 6,
        "stackwise_depth_coefficients": [1.1] * 6,
        "stackwise_kernel_sizes": [3, 3, 3, 5, 5, 5],
        "stackwise_num_repeats": [1, 2, 4, 5, 4, 2],
        "stackwise_input_filters": [32, 24, 32, 48, 96, 144],
        "stackwise_output_filters": [24, 32, 48, 96, 144, 192],
        "stackwise_expansion_ratios": [4, 8, 8, 8, 8, 8],
        "stackwise_strides": [1, 2, 2, 2, 1, 2],
        "stackwise_squeeze_and_excite_ratios": [0] * 6,
        "stackwise_block_types": ["fused"] * 3 + ["unfused"] * 3,
        "stackwise_force_input_filters": [24, 0, 0, 0, 0, 0],
        "stackwise_nores_option": [True] + [False] * 5,
        "activation": "relu",
    },
    "es": {
        "stackwise_width_coefficients": [1.0] * 6,
        "stackwise_depth_coefficients": [1.0] * 6,
        "stackwise_kernel_sizes": [3, 3, 3, 5, 5, 5],
        "stackwise_num_repeats": [1, 2, 4, 5, 4, 2],
        "stackwise_input_filters": [32, 24, 32, 48, 96, 144],
        "stackwise_output_filters": [24, 32, 48, 96, 144, 192],
        "stackwise_expansion_ratios": [4, 8, 8, 8, 8, 8],
        "stackwise_strides": [1, 2, 2, 2, 1, 2],
        "stackwise_squeeze_and_excite_ratios": [0] * 6,
        "stackwise_block_types": ["fused"] * 3 + ["unfused"] * 3,
        "stackwise_force_input_filters": [24, 0, 0, 0, 0, 0],
        "stackwise_nores_option": [True] + [False] * 5,
        "activation": "relu",
    },
    "rw_m": {
        "stackwise_width_coefficients": [1.2] * 6,
        "stackwise_depth_coefficients": [1.2] * 4 + [1.6] * 2,
        "stackwise_kernel_sizes": [3, 3, 3, 3, 3, 3],
        "stackwise_num_repeats": [2, 4, 4, 6, 9, 15],
        "stackwise_input_filters": [24, 24, 48, 64, 128, 160],
        "stackwise_output_filters": [24, 48, 64, 128, 160, 272],
        "stackwise_expansion_ratios": [1, 4, 4, 4, 6, 6],
        "stackwise_strides": [1, 2, 2, 2, 1, 2],
        "stackwise_squeeze_and_excite_ratios": [0, 0, 0, 0.25, 0.25, 0.25],
        "stackwise_block_types": ["fused"] * 3 + ["unfused"] * 3,
        "stackwise_force_input_filters": [0, 0, 0, 0, 0, 0],
        "stackwise_nores_option": [False] * 6,
        "activation": "silu",
        "num_features": 1792,
    },
    "rw_s": {
        "stackwise_width_coefficients": [1.0] * 6,
        "stackwise_depth_coefficients": [1.0] * 6,
        "stackwise_kernel_sizes": [3, 3, 3, 3, 3, 3],
        "stackwise_num_repeats": [2, 4, 4, 6, 9, 15],
        "stackwise_input_filters": [24, 24, 48, 64, 128, 160],
        "stackwise_output_filters": [24, 48, 64, 128, 160, 272],
        "stackwise_expansion_ratios": [1, 4, 4, 4, 6, 6],
        "stackwise_strides": [1, 2, 2, 2, 1, 2],
        "stackwise_squeeze_and_excite_ratios": [0, 0, 0, 0.25, 0.25, 0.25],
        "stackwise_block_types": ["fused"] * 3 + ["unfused"] * 3,
        "stackwise_force_input_filters": [0, 0, 0, 0, 0, 0],
        "stackwise_nores_option": [False] * 6,
        "activation": "silu",
        "num_features": 1792,
    },
    "rw_t": {
        "stackwise_width_coefficients": [0.8] * 6,
        "stackwise_depth_coefficients": [0.9] * 6,
        "stackwise_kernel_sizes": [3, 3, 3, 3, 3, 3],
        "stackwise_num_repeats": [2, 4, 4, 6, 9, 15],
        "stackwise_input_filters": [24, 24, 48, 64, 128, 160],
        "stackwise_output_filters": [24, 48, 64, 128, 160, 256],
        "stackwise_expansion_ratios": [1, 4, 4, 4, 6, 6],
        "stackwise_strides": [1, 2, 2, 2, 1, 2],
        "stackwise_squeeze_and_excite_ratios": [0, 0, 0, 0.25, 0.25, 0.25],
        "stackwise_block_types": ["cba"] + ["fused"] * 2 + ["unfused"] * 3,
        "stackwise_force_input_filters": [0, 0, 0, 0, 0, 0],
        "stackwise_nores_option": [False] * 6,
        "activation": "silu",
    },
}


def convert_backbone_config(timm_config):
    timm_architecture = timm_config["architecture"]

    base_kwargs = {
        "stackwise_kernel_sizes": [3, 3, 5, 3, 5, 5, 3],
        "stackwise_num_repeats": [1, 2, 2, 3, 3, 4, 1],
        "stackwise_input_filters": [32, 16, 24, 40, 80, 112, 192],
        "stackwise_output_filters": [16, 24, 40, 80, 112, 192, 320],
        "stackwise_expansion_ratios": [1, 6, 6, 6, 6, 6, 6],
        "stackwise_strides": [1, 2, 2, 2, 1, 2, 1],
        "stackwise_block_types": ["v1"] * 7,
        "min_depth": None,
        "include_stem_padding": True,
        "use_depth_divisor_as_min_depth": True,
        "cap_round_filter_decrease": True,
        "stem_conv_padding": "valid",
        "batch_norm_momentum": 0.9,
        "batch_norm_epsilon": 1e-5,
        "dropout": 0,
        "projection_activation": None,
    }

    variant = "_".join(timm_architecture.split("_")[1:])

    if variant not in VARIANT_MAP:
        raise ValueError(
            f"Currently, the architecture {timm_architecture} is not supported."
        )

    base_kwargs.update(VARIANT_MAP[variant])

    return base_kwargs


def convert_weights(backbone, loader, timm_config):
    timm_architecture = timm_config["architecture"]
    variant = "_".join(timm_architecture.split("_")[1:])

    def port_conv2d(keras_layer, hf_weight_prefix, port_bias=True):
        loader.port_weight(
            keras_layer.kernel,
            hf_weight_key=f"{hf_weight_prefix}.weight",
            hook_fn=lambda x, _: np.transpose(x, (2, 3, 1, 0)),
        )

        if port_bias:
            loader.port_weight(
                keras_layer.bias,
                hf_weight_key=f"{hf_weight_prefix}.bias",
            )

    def port_depthwise_conv2d(
        keras_layer,
        hf_weight_prefix,
        port_bias=True,
        depth_multiplier=1,
    ):
        def convert_pt_conv2d_kernel(pt_kernel):
            out_channels, in_channels_per_group, height, width = pt_kernel.shape
            # PT Convs are depthwise convs if and only if
            # `in_channels_per_group == 1`
            assert in_channels_per_group == 1
            pt_kernel = np.transpose(pt_kernel, (2, 3, 0, 1))
            in_channels = out_channels // depth_multiplier
            return np.reshape(
                pt_kernel, (height, width, in_channels, depth_multiplier)
            )

        loader.port_weight(
            keras_layer.kernel,
            hf_weight_key=f"{hf_weight_prefix}.weight",
            hook_fn=lambda x, _: convert_pt_conv2d_kernel(x),
        )

        if port_bias:
            loader.port_weight(
                keras_layer.bias,
                hf_weight_key=f"{hf_weight_prefix}.bias",
            )

    def port_batch_normalization(keras_layer, hf_weight_prefix):
        loader.port_weight(
            keras_layer.gamma,
            hf_weight_key=f"{hf_weight_prefix}.weight",
        )
        loader.port_weight(
            keras_layer.beta,
            hf_weight_key=f"{hf_weight_prefix}.bias",
        )
        loader.port_weight(
            keras_layer.moving_mean,
            hf_weight_key=f"{hf_weight_prefix}.running_mean",
        )
        loader.port_weight(
            keras_layer.moving_variance,
            hf_weight_key=f"{hf_weight_prefix}.running_var",
        )
        # do we need num batches tracked?

    # Stem
    port_conv2d(backbone.get_layer("stem_conv"), "conv_stem", port_bias=False)
    port_batch_normalization(backbone.get_layer("stem_bn"), "bn1")

    # Stages
    num_stacks = len(backbone.stackwise_kernel_sizes)

    for stack_index in range(num_stacks):
        block_type = backbone.stackwise_block_types[stack_index]
        expansion_ratio = backbone.stackwise_expansion_ratios[stack_index]
        repeats = backbone.stackwise_num_repeats[stack_index]
        stack_depth_coefficient = backbone.stackwise_depth_coefficients[
            stack_index
        ]

        repeats = int(math.ceil(stack_depth_coefficient * repeats))

        se_ratio = VARIANT_MAP[variant]["stackwise_squeeze_and_excite_ratios"][
            stack_index
        ]

        for block_idx in range(repeats):
            conv_pw_count = 0
            bn_count = 1

            # 97 is the start of the lowercase alphabet.
            letter_identifier = chr(block_idx + 97)

            keras_block_prefix = f"block{stack_index + 1}{letter_identifier}_"
            hf_block_prefix = f"blocks.{stack_index}.{block_idx}."

            if block_type == "v1":
                conv_pw_name_map = ["conv_pw", "conv_pwl"]
                # Initial Expansion Conv
                if expansion_ratio != 1:
                    port_conv2d(
                        backbone.get_layer(keras_block_prefix + "expand_conv"),
                        hf_block_prefix + conv_pw_name_map[conv_pw_count],
                        port_bias=False,
                    )
                    conv_pw_count += 1
                    port_batch_normalization(
                        backbone.get_layer(keras_block_prefix + "expand_bn"),
                        hf_block_prefix + f"bn{bn_count}",
                    )
                    bn_count += 1

                # Depthwise Conv
                port_depthwise_conv2d(
                    backbone.get_layer(keras_block_prefix + "dwconv"),
                    hf_block_prefix + "conv_dw",
                    port_bias=False,
                )
                port_batch_normalization(
                    backbone.get_layer(keras_block_prefix + "dwconv_bn"),
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1

                if 0 < se_ratio <= 1:
                    # Squeeze and Excite
                    port_conv2d(
                        backbone.get_layer(keras_block_prefix + "se_reduce"),
                        hf_block_prefix + "se.conv_reduce",
                    )
                    port_conv2d(
                        backbone.get_layer(keras_block_prefix + "se_expand"),
                        hf_block_prefix + "se.conv_expand",
                    )

                # Output/Projection
                port_conv2d(
                    backbone.get_layer(keras_block_prefix + "project"),
                    hf_block_prefix + conv_pw_name_map[conv_pw_count],
                    port_bias=False,
                )
                conv_pw_count += 1
                port_batch_normalization(
                    backbone.get_layer(keras_block_prefix + "project_bn"),
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1
            elif block_type == "fused":
                fused_block_layer = backbone.get_layer(keras_block_prefix)

                # Initial Expansion Conv
                port_conv2d(
                    fused_block_layer.conv1,
                    hf_block_prefix + "conv_exp",
                    port_bias=False,
                )
                conv_pw_count += 1
                port_batch_normalization(
                    fused_block_layer.bn1,
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1

                if 0 < se_ratio <= 1:
                    # Squeeze and Excite
                    port_conv2d(
                        fused_block_layer.se_conv1,
                        hf_block_prefix + "se.conv_reduce",
                    )
                    port_conv2d(
                        fused_block_layer.se_conv2,
                        hf_block_prefix + "se.conv_expand",
                    )

                # Output/Projection
                port_conv2d(
                    fused_block_layer.output_conv,
                    hf_block_prefix + "conv_pwl",
                    port_bias=False,
                )
                conv_pw_count += 1
                port_batch_normalization(
                    fused_block_layer.bn2,
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1

            elif block_type == "unfused":
                unfused_block_layer = backbone.get_layer(keras_block_prefix)
                # Initial Expansion Conv
                if expansion_ratio != 1:
                    port_conv2d(
                        unfused_block_layer.conv1,
                        hf_block_prefix + "conv_pw",
                        port_bias=False,
                    )
                    conv_pw_count += 1
                    port_batch_normalization(
                        unfused_block_layer.bn1,
                        hf_block_prefix + f"bn{bn_count}",
                    )
                    bn_count += 1

                # Depthwise Conv
                port_depthwise_conv2d(
                    unfused_block_layer.depthwise,
                    hf_block_prefix + "conv_dw",
                    port_bias=False,
                )
                port_batch_normalization(
                    unfused_block_layer.bn2,
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1

                if 0 < se_ratio <= 1:
                    # Squeeze and Excite
                    port_conv2d(
                        unfused_block_layer.se_conv1,
                        hf_block_prefix + "se.conv_reduce",
                    )
                    port_conv2d(
                        unfused_block_layer.se_conv2,
                        hf_block_prefix + "se.conv_expand",
                    )

                # Output/Projection
                port_conv2d(
                    unfused_block_layer.output_conv,
                    hf_block_prefix + "conv_pwl",
                    port_bias=False,
                )
                conv_pw_count += 1
                port_batch_normalization(
                    unfused_block_layer.bn3,
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1
            elif block_type == "cba":
                cba_block_layer = backbone.get_layer(keras_block_prefix)
                # Initial Expansion Conv
                port_conv2d(
                    cba_block_layer.conv1,
                    hf_block_prefix + "conv",
                    port_bias=False,
                )
                conv_pw_count += 1
                port_batch_normalization(
                    cba_block_layer.bn1,
                    hf_block_prefix + f"bn{bn_count}",
                )
                bn_count += 1

    # Head/Top
    port_conv2d(backbone.get_layer("top_conv"), "conv_head", port_bias=False)
    port_batch_normalization(backbone.get_layer("top_bn"), "bn2")


def convert_head(task, loader, timm_config):
    classifier_prefix = timm_config["pretrained_cfg"]["classifier"]
    prefix = f"{classifier_prefix}."
    loader.port_weight(
        task.output_dense.kernel,
        hf_weight_key=prefix + "weight",
        hook_fn=lambda x, _: np.transpose(np.squeeze(x)),
    )
    loader.port_weight(
        task.output_dense.bias,
        hf_weight_key=prefix + "bias",
    )



from keras_hub.src.models.resnet.resnet_backbone import ResNetBackbone
from keras_hub.src.models.resnet.resnet_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, ResNetBackbone)



import numpy as np

from keras_hub.src.models.distil_bert.distil_bert_backbone import (
    DistilBertBackbone,
)
from keras_hub.src.utils.preset_utils import HF_TOKENIZER_CONFIG_FILE
from keras_hub.src.utils.preset_utils import get_file
from keras_hub.src.utils.preset_utils import load_json

backbone_cls = DistilBertBackbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "num_layers": transformers_config["n_layers"],
        "num_heads": transformers_config["n_heads"],
        "hidden_dim": transformers_config["dim"],
        "intermediate_dim": transformers_config["hidden_dim"],
        "dropout": transformers_config["dropout"],
        "max_sequence_length": transformers_config["max_position_embeddings"],
    }


def convert_weights(backbone, loader, transformers_config):
    # Embeddings
    loader.port_weight(
        keras_variable=backbone.get_layer(
            "token_and_position_embedding"
        ).token_embedding.embeddings,
        hf_weight_key="distilbert.embeddings.word_embeddings.weight",
    )
    loader.port_weight(
        keras_variable=backbone.get_layer(
            "token_and_position_embedding"
        ).position_embedding.position_embeddings,
        hf_weight_key="distilbert.embeddings.position_embeddings.weight",
    )

    # Attention blocks
    for index in range(backbone.num_layers):
        decoder_layer = backbone.transformer_layers[index]

        # Norm layers
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer_norm.gamma,
            hf_weight_key=f"distilbert.transformer.layer.{index}.sa_layer_norm.weight",
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer_norm.beta,
            hf_weight_key=f"distilbert.transformer.layer.{index}.sa_layer_norm.bias",
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_layer_norm.gamma,
            hf_weight_key=f"distilbert.transformer.layer.{index}.output_layer_norm.weight",
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_layer_norm.beta,
            hf_weight_key=f"distilbert.transformer.layer.{index}.output_layer_norm.bias",
        )

        # Attention layers
        # Query
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.query_dense.kernel,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.q_lin.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                np.transpose(hf_tensor), keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.query_dense.bias,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.q_lin.bias",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor, keras_shape
            ),
        )

        # Key
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.key_dense.kernel,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.k_lin.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                np.transpose(hf_tensor), keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.key_dense.bias,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.k_lin.bias",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor, keras_shape
            ),
        )

        # Value
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.value_dense.kernel,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.v_lin.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                np.transpose(hf_tensor), keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.value_dense.bias,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.v_lin.bias",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor, keras_shape
            ),
        )

        # Output
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.output_dense.kernel,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.out_lin.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                np.transpose(hf_tensor), keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.output_dense.bias,
            hf_weight_key=f"distilbert.transformer.layer.{index}.attention.out_lin.bias",
        )

        # MLP layers
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_intermediate_dense.kernel,
            hf_weight_key=f"distilbert.transformer.layer.{index}.ffn.lin1.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_intermediate_dense.bias,
            hf_weight_key=f"distilbert.transformer.layer.{index}.ffn.lin1.bias",
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_output_dense.kernel,
            hf_weight_key=f"distilbert.transformer.layer.{index}.ffn.lin2.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_output_dense.bias,
            hf_weight_key=f"distilbert.transformer.layer.{index}.ffn.lin2.bias",
        )

    # Normalization
    loader.port_weight(
        keras_variable=backbone.embeddings_layer_norm.gamma,
        hf_weight_key="distilbert.embeddings.LayerNorm.weight",
    )
    loader.port_weight(
        keras_variable=backbone.embeddings_layer_norm.beta,
        hf_weight_key="distilbert.embeddings.LayerNorm.bias",
    )


def convert_tokenizer(cls, preset, **kwargs):
    transformers_config = load_json(preset, HF_TOKENIZER_CONFIG_FILE)
    return cls(
        get_file(preset, "vocab.txt"),
        lowercase=transformers_config["do_lower_case"],
        **kwargs,
    )



from keras_hub.src.models.xlm_roberta.xlm_roberta_backbone import (
    XLMRobertaBackbone,
)
from keras_hub.src.models.xlm_roberta.xlm_roberta_presets import (
    backbone_presets,
)
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, XLMRobertaBackbone)



from keras_hub.src.models.whisper.whisper_backbone import WhisperBackbone
from keras_hub.src.models.whisper.whisper_presets import backbone_presets
from keras_hub.src.utils.preset_utils import register_presets

register_presets(backbone_presets, WhisperBackbone)



from keras_hub.src.models.seq_2_seq_lm_preprocessor import Seq2SeqLMPreprocessor


class AudioToTextPreprocessor(Seq2SeqLMPreprocessor):
    """Base class for audio-to-text preprocessing layers.

    `AudioToTextPreprocessor` layers wrap an audio feature extractor (specific
    to the subclass) and a `keras_hub.tokenizer.Tokenizer` to create a
    preprocessing layer for audio-to-text tasks. It is intended to be
    paired with a `keras_hub.models.AudioToText` task.

    Subclasses are expected to handle the conversion of raw audio data into
    numerical features suitable for an encoder, and raw text data into token IDs
    for a decoder.

    All `AudioToTextPreprocessor` layers take a dictionary as input,
    typically with keys like `"audio"` (for audio data) and `"text"` (for
    target transcriptions or decoder prompts).

    This layer will always output a `(x, y, sample_weight)` tuple, where `x`
    is a dictionary containing processed audio features for the encoder and
    tokenized text inputs for the decoder. `y` contains the target token IDs
    (decoder input tokens shifted by one position), and `sample_weight`
    indicates padding in `y`. The exact keys and structure of features within
    `x` will depend on the specific subclass and the paired `AudioToText` model.

    An `AudioToTextPreprocessor` includes `generate_preprocess` and
    `generate_postprocess` methods for use during inference with an
    `AudioToText` model's `generate()` method.

    All `AudioToTextPreprocessor` tasks include a `from_preset()` constructor
    which can be used to load a pre-trained configuration, including tokenizer
    vocabularies and audio feature extraction settings. Calling `from_preset()`
    on this base class can instantiate the correct subclass registered for the
    given preset.

    Examples:
    ```python
    preprocessor = keras_hub.models.AudioToTextPreprocessor.from_preset(
        "moonshine_base_en",
        decoder_sequence_length=10
    )

    # Process a single audio-text pair.
    x = {
        "audio": keras.random.normal((1, 16000, 1)),
        "text": ["the quick brown fox"]
    }
    x, y, sample_weight = preprocessor(x)

    # Process a batch of audio-text pairs.
    x = {
        "audio": keras.random.normal((2, 16000, 1)),
        "text": ["first sentence", "second sentence"]
    }
    x, y, sample_weight = preprocessor(x)

    # With a `tf.data.Dataset`.
    audio_tf = keras.ops.convert_to_tensor(batch_input["audio"])
    text_tf = batch_input["text"] # List of strings
    x = {"audio": audio_tf, "text": text_tf}
    ds = tf.data.Dataset.from_tensor_slices(x)
    ds = ds.map(preprocessor, num_parallel_calls=tf.data.AUTOTUNE)
    ds = ds.batch(2) # Batching after map

    # Generate preprocess and postprocess.
    x = preprocessor.generate_preprocess({
        "audio": keras.random.normal((1, 16000, 1)),
        "text": ["optional prompt text"]
    })
    x = preprocessor.generate_postprocess({
        "decoder_token_ids": keras.ops.array([[10, 20, 30, 2, 0]]),
        "decoder_padding_mask": keras.ops.array([
            [True, True, True, True, False]
        ])
    })
    ```
    """

    # TODO: Fill in once audio to text task model requirements are clearer.



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.siglip.siglip_backbone import SigLIPBackbone
from keras_hub.src.tokenizers.sentence_piece_tokenizer import (
    SentencePieceTokenizer,
)


@keras_hub_export(
    [
        "keras_hub.tokenizers.SigLIPTokenizer",
        "keras_hub.models.SigLIPTokenizer",
    ]
)
class SigLIPTokenizer(SentencePieceTokenizer):
    """SigLIP tokenizer layer based on SentencePiece.

    This tokenizer class will tokenize raw strings into integer sequences and
    is based on `keras_hub.tokenizers.SentencePieceTokenizer`. Unlike the
    underlying tokenizer, it will check for all special tokens needed by
    SigLIP models and provides a `from_preset()` method to automatically
    download a matching vocabulary for a SigLIP preset.

    If input is a batch of strings (rank > 0), the layer will output a
    `tf.RaggedTensor` where the last dimension of the output is ragged.

    If input is a scalar string (rank == 0), the layer will output a dense
    `tf.Tensor` with static shape `[None]`.

    Args:
        proto: Either a `string` path to a SentencePiece proto file, or a
            `bytes` object with a serialized SentencePiece proto. See the
            [SentencePiece repository](https://github.com/google/sentencepiece)
            for more details on the format.

    Examples:

    ```python
    # Unbatched input.
    tokenizer = keras_hub.models.SigLIPTokenizer.from_preset(
        "siglip_base_patch16_224"
    )
    tokenizer("The quick brown fox jumped.")

    # Batched input.
    tokenizer(["The quick brown fox jumped.", "The fox slept."])

    # Detokenization.
    tokenizer.detokenize(tokenizer("The quick brown fox jumped."))

    # Custom vocabulary.
    bytes_io = io.BytesIO()
    ds = tf.data.Dataset.from_tensor_slices(["The quick brown fox jumped."])
    sentencepiece.SentencePieceTrainer.train(
        sentence_iterator=ds.as_numpy_iterator(),
        model_writer=bytes_io,
        vocab_size=8,
        model_type="WORD",
        pad_id=0,
        bos_id=1,
        eos_id=2,
        unk_id=3,
        unk_piece="<unk>",
    )
    tokenizer = keras_hub.models.SigLIPTokenizer(
        proto=bytes_io.getvalue(),
    )
    tokenizer("The quick brown fox jumped.")
    ```
    """

    backbone_cls = SigLIPBackbone

    def __init__(self, proto, **kwargs):
        self._add_special_token("<unk>", "unknown_token")
        self._add_special_token("</s>", "end_token")
        self._add_special_token("</s>", "pad_token")
        super().__init__(proto=proto, **kwargs)



import numpy as np

from keras_hub.src.models.gpt2.gpt2_backbone import GPT2Backbone
from keras_hub.src.utils.preset_utils import get_file

backbone_cls = GPT2Backbone


def convert_backbone_config(transformers_config):
    return {
        "vocabulary_size": transformers_config["vocab_size"],
        "num_layers": transformers_config["n_layer"],
        "num_heads": transformers_config["n_head"],
        "hidden_dim": transformers_config["n_embd"],
        "intermediate_dim": transformers_config["n_embd"] * 4,
        "dropout": transformers_config["resid_pdrop"],
        "max_sequence_length": transformers_config["n_positions"],
    }


def convert_weights(backbone, loader, transformers_config):
    # Embeddings
    loader.port_weight(
        keras_variable=backbone.token_embedding.embeddings,
        hf_weight_key="wte.weight",
    )
    loader.port_weight(
        keras_variable=backbone.position_embedding.position_embeddings,
        hf_weight_key="wpe.weight",
    )

    # Attention blocks
    for index in range(backbone.num_layers):
        decoder_layer = backbone.transformer_layers[index]

        # Norm layers
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer_norm.gamma,
            hf_weight_key=f"h.{index}.ln_1.weight",
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer_norm.beta,
            hf_weight_key=f"h.{index}.ln_1.bias",
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_layer_norm.gamma,
            hf_weight_key=f"h.{index}.ln_2.weight",
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_layer_norm.beta,
            hf_weight_key=f"h.{index}.ln_2.bias",
        )

        # Attention layers
        n_embd = transformers_config["n_embd"]

        # Query
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.query_dense.kernel,
            hf_weight_key=f"h.{index}.attn.c_attn.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor[:, :n_embd], keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.query_dense.bias,
            hf_weight_key=f"h.{index}.attn.c_attn.bias",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor[:n_embd], keras_shape
            ),
        )

        # Key
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.key_dense.kernel,
            hf_weight_key=f"h.{index}.attn.c_attn.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor[:, n_embd : 2 * n_embd], keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.key_dense.bias,
            hf_weight_key=f"h.{index}.attn.c_attn.bias",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor[n_embd : 2 * n_embd], keras_shape
            ),
        )

        # Value
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.value_dense.kernel,
            hf_weight_key=f"h.{index}.attn.c_attn.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor[:, 2 * n_embd :], keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.value_dense.bias,
            hf_weight_key=f"h.{index}.attn.c_attn.bias",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor[2 * n_embd :], keras_shape
            ),
        )

        # Output
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.output_dense.kernel,
            hf_weight_key=f"h.{index}.attn.c_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor, keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._self_attention_layer.output_dense.bias,
            hf_weight_key=f"h.{index}.attn.c_proj.bias",
        )

        # MLP layers
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_intermediate_dense.kernel,
            hf_weight_key=f"h.{index}.mlp.c_fc.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor, keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_intermediate_dense.bias,
            hf_weight_key=f"h.{index}.mlp.c_fc.bias",
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_output_dense.kernel,
            hf_weight_key=f"h.{index}.mlp.c_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.reshape(
                hf_tensor, keras_shape
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer._feedforward_output_dense.bias,
            hf_weight_key=f"h.{index}.mlp.c_proj.bias",
        )

    # Normalization
    loader.port_weight(
        keras_variable=backbone.layer_norm.gamma,
        hf_weight_key="ln_f.weight",
    )
    loader.port_weight(
        keras_variable=backbone.layer_norm.beta,
        hf_weight_key="ln_f.bias",
    )


def convert_tokenizer(cls, preset, **kwargs):
    vocab_file = get_file(preset, "vocab.json")
    merges_file = get_file(preset, "merges.txt")
    return cls(
        vocabulary=vocab_file,
        merges=merges_file,
        **kwargs,
    )



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.llama.llama_causal_lm import LlamaCausalLM
from keras_hub.src.models.llama3.llama3_backbone import Llama3Backbone
from keras_hub.src.models.llama3.llama3_causal_lm_preprocessor import (
    Llama3CausalLMPreprocessor,
)


@keras_hub_export("keras_hub.models.Llama3CausalLM")
class Llama3CausalLM(LlamaCausalLM):
    """An end-to-end Llama 3 model for causal language modeling.

    A causal language model (LM) predicts the next token based on previous
    tokens. This task setup can be used to train the model unsupervised on
    plain text input, or to autoregressively generate plain text similar to
    the data used for training. This task can be used for pre-training or
    fine-tuning a LLaMA 3 model, simply by calling `fit()`.

    This model has a `generate()` method, which generates text based on a
    prompt. The generation strategy used is controlled by an additional
    `sampler` argument on `compile()`. You can recompile the model with
    different `keras_hub.samplers` objects to control the generation. By
    default, `"top_k"` sampling will be used.

    Args:
        backbone: A `keras_hub.models.Llama3Backbone` instance.
        preprocessor: A `keras_hub.models.Llama3CausalLMPreprocessor` or `None`.
            If `None`, this model will not apply preprocessing, and inputs
            should be preprocessed before calling the model.
    """

    backbone_cls = Llama3Backbone
    preprocessor_cls = Llama3CausalLMPreprocessor



from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.image_classifier_preprocessor import (
    ImageClassifierPreprocessor,
)
from keras_hub.src.models.mobilenet.mobilenet_backbone import MobileNetBackbone
from keras_hub.src.models.mobilenet.mobilenet_image_converter import (
    MobileNetImageConverter,
)


@keras_hub_export("keras_hub.models.MobileNetImageClassifierPreprocessor")
class MobileNetImageClassifierPreprocessor(ImageClassifierPreprocessor):
    backbone_cls = MobileNetBackbone
    image_converter_cls = MobileNetImageConverter



from keras_hub.src.api_export import keras_hub_export

# Unique source of truth for the version number.
__version__ = "0.21.0"


@keras_hub_export("keras_hub.version")
def version():
    return __version__



import keras

from keras_hub.src.api_export import keras_hub_export
from keras_hub.src.models.backbone import Backbone
from keras_hub.src.models.flux.flux_layers import DoubleStreamBlock
from keras_hub.src.models.flux.flux_layers import EmbedND
from keras_hub.src.models.flux.flux_layers import LastLayer
from keras_hub.src.models.flux.flux_layers import MLPEmbedder
from keras_hub.src.models.flux.flux_layers import SingleStreamBlock
from keras_hub.src.models.flux.flux_maths import TimestepEmbedding


@keras_hub_export("keras_hub.models.FluxBackbone")
class FluxBackbone(Backbone):
    """Transformer model for flow matching on sequences.

    The model processes image and text data with associated positional and
    timestep embeddings, and optionally applies guidance embedding.
    Double-stream blocks handle separate image and text streams, while
    single-stream blocks combine these streams. Ported from:
    https://github.com/black-forest-labs/flux


    Args:
        input_channels: int. The number of input channels.
        hidden_size: int. The hidden size of the transformer, must be divisible
            by `num_heads`.
        mlp_ratio: float. The ratio of the MLP dimension to the hidden size.
        num_heads: int. The number of attention heads.
        depth: int. The number of double-stream blocks.
        depth_single_blocks: int. The number of single-stream blocks.
        axes_dim: list[int]. A list of dimensions for the positional embedding
            axes.
        theta: int. The base frequency for positional embeddings.
        use_bias: bool. Whether to apply bias to the query, key, and value
            projections.
        guidance_embed: bool. If True, applies guidance embedding in the model.

    Call arguments:
        image: KerasTensor. Image input tensor of shape (N, L, D) where N is the
            batch size, L is the sequence length, and D is the feature
            dimension.
        image_ids: KerasTensor. Image ID input tensor of shape (N, L, D)
            corresponding to the image sequences.
        text: KerasTensor. Text input tensor of shape (N, L, D).
        text_ids: KerasTensor. Text ID input tensor of shape (N, L, D)
            corresponding to the text sequences.
        timesteps: KerasTensor. Timestep tensor used to compute positional
            embeddings.
        y: KerasTensor. Additional vector input, such as target values.
        guidance: KerasTensor, optional. Guidance input tensor used
            in guidance-embedded models.
    Raises:
        ValueError: If `hidden_size` is not divisible by `num_heads`, or if
            `sum(axes_dim)` is not equal to the positional embedding dimension.
    """

    def __init__(
        self,
        input_channels,
        hidden_size,
        mlp_ratio,
        num_heads,
        depth,
        depth_single_blocks,
        axes_dim,
        theta,
        use_bias,
        guidance_embed=False,
        # These will be inferred from the CLIP/T5 encoders later
        image_shape=(None, 768, 3072),
        text_shape=(None, 768, 3072),
        image_ids_shape=(None, 768, 3072),
        text_ids_shape=(None, 768, 3072),
        y_shape=(None, 128),
        **kwargs,
    ):
        # === Layers ===
        self.positional_embedder = EmbedND(theta=theta, axes_dim=axes_dim)
        self.image_input_embedder = keras.layers.Dense(
            hidden_size, use_bias=True
        )
        self.time_input_embedder = MLPEmbedder(hidden_dim=hidden_size)
        self.vector_embedder = MLPEmbedder(hidden_dim=hidden_size)
        self.guidance_input_embedder = (
            MLPEmbedder(hidden_dim=hidden_size)
            if guidance_embed
            else keras.layers.Identity()
        )
        self.text_input_embedder = keras.layers.Dense(hidden_size)

        self.double_blocks = [
            DoubleStreamBlock(
                hidden_size,
                num_heads,
                mlp_ratio=mlp_ratio,
                use_bias=use_bias,
            )
            for _ in range(depth)
        ]

        self.single_blocks = [
            SingleStreamBlock(hidden_size, num_heads, mlp_ratio=mlp_ratio)
            for _ in range(depth_single_blocks)
        ]

        self.final_layer = LastLayer(hidden_size, 1, input_channels)
        self.timestep_embedding = TimestepEmbedding()
        self.guidance_embed = guidance_embed

        # === Functional Model ===
        image_input = keras.Input(shape=image_shape, name="image")
        image_ids = keras.Input(shape=image_ids_shape, name="image_ids")
        text_input = keras.Input(shape=text_shape, name="text")
        text_ids = keras.Input(shape=text_ids_shape, name="text_ids")
        y = keras.Input(shape=y_shape, name="y")
        timesteps_input = keras.Input(shape=(), name="timesteps")
        guidance_input = keras.Input(shape=(), name="guidance")

        # running on sequences image
        image = self.image_input_embedder(image_input)
        modulation_encoding = self.time_input_embedder(
            self.timestep_embedding(timesteps_input, dim=256)
        )
        if self.guidance_embed:
            if guidance_input is None:
                raise ValueError(
                    "Didn't get guidance strength for guidance distilled model."
                )
            modulation_encoding = (
                modulation_encoding
                + self.guidance_input_embedder(
                    self.timestep_embedding(guidance_input, dim=256)
                )
            )

        modulation_encoding = modulation_encoding + self.vector_embedder(y)
        text = self.text_input_embedder(text_input)

        ids = keras.ops.concatenate((text_ids, image_ids), axis=1)
        positional_encoding = self.positional_embedder(ids)

        for block in self.double_blocks:
            image, text = block(
                image=image,
                text=text,
                modulation_encoding=modulation_encoding,
                positional_encoding=positional_encoding,
            )

        image = keras.ops.concatenate((text, image), axis=1)
        for block in self.single_blocks:
            image = block(
                image,
                modulation_encoding=modulation_encoding,
                positional_encoding=positional_encoding,
            )
        image = image[:, text.shape[1] :, ...]

        image = self.final_layer(
            image, modulation_encoding
        )  # (N, T, patch_size ** 2 * output_channels)

        super().__init__(
            inputs={
                "image": image_input,
                "image_ids": image_ids,
                "text": text_input,
                "text_ids": text_ids,
                "y": y,
                "timesteps": timesteps_input,
                "guidance": guidance_input,
            },
            outputs=image,
            **kwargs,
        )

        # === Config ===
        self.input_channels = input_channels
        self.output_channels = self.input_channels
        self.hidden_size = hidden_size
        self.num_heads = num_heads
        self.image_shape = image_shape
        self.text_shape = text_shape
        self.image_ids_shape = image_ids_shape
        self.text_ids_shape = text_ids_shape
        self.y_shape = y_shape
        self.mlp_ratio = mlp_ratio
        self.depth = depth
        self.depth_single_blocks = depth_single_blocks
        self.axes_dim = axes_dim
        self.theta = theta
        self.use_bias = use_bias

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "input_channels": self.input_channels,
                "hidden_size": self.hidden_size,
                "mlp_ratio": self.mlp_ratio,
                "num_heads": self.num_heads,
                "depth": self.depth,
                "depth_single_blocks": self.depth_single_blocks,
                "axes_dim": self.axes_dim,
                "theta": self.theta,
                "use_bias": self.use_bias,
                "guidance_embed": self.guidance_embed,
                "image_shape": self.image_shape,
                "text_shape": self.text_shape,
                "image_ids_shape": self.image_ids_shape,
                "text_ids_shape": self.text_ids_shape,
                "y_shape": self.y_shape,
            }
        )
        return config

    def encode_text_step(self, token_ids, negative_token_ids):
        raise NotImplementedError("Not implemented yet")

        def encode(token_ids):
            raise NotImplementedError("Not implemented yet")

    def encode_image_step(self, images):
        raise NotImplementedError("Not implemented yet")

    def add_noise_step(self, latents, noises, step, num_steps):
        raise NotImplementedError("Not implemented yet")

    def denoise_step(
        self,
    ):
        raise NotImplementedError("Not implemented yet")

    def decode_step(self, latents):
        raise NotImplementedError("Not implemented yet")



"""RetinaNet model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "retinanet_resnet50_fpn_coco": {
        "metadata": {
            "description": (
                "RetinaNet model with ResNet50 backbone fine-tuned on COCO in "
                "800x800 resolution."
            ),
            "params": 34121239,
            "path": "retinanet",
        },
        "kaggle_handle": "kaggle://keras/retinanet/keras/retinanet_resnet50_fpn_coco/4",
    },
    "retinanet_resnet50_fpn_v2_coco": {
        "metadata": {
            "description": (
                "RetinaNet model with ResNet50 backbone fine-tuned on COCO in "
                "800x800 resolution with FPN features created from P5 level."
            ),
            "params": 31558592,
            "path": "retinanet",
        },
        "kaggle_handle": "kaggle://keras/retinanet/keras/retinanet_resnet50_fpn_v2_coco/3",
    },
}



"""OPT model preset configurations."""

# Metadata for loading pretrained model weights.
backbone_presets = {
    "opt_125m_en": {
        "metadata": {
            "description": (
                "12-layer OPT model where case in maintained. Trained on "
                "BookCorpus, CommonCrawl, Pile, and PushShift.io corpora."
            ),
            "params": 125237760,
            "path": "opt",
        },
        "kaggle_handle": "kaggle://keras/opt/keras/opt_125m_en/3",
    },
    # We skip the 350m checkpoint because it does not match the structure of
    # other checkpoints.
    "opt_1.3b_en": {
        "metadata": {
            "description": (
                "24-layer OPT model where case in maintained. Trained on "
                "BookCorpus, CommonCrawl, Pile, and PushShift.io corpora."
            ),
            "params": 1315753984,
            "path": "opt",
        },
        "kaggle_handle": "kaggle://keras/opt/keras/opt_1.3b_en/3",
    },
    "opt_2.7b_en": {
        "metadata": {
            "description": (
                "32-layer OPT model where case in maintained. Trained on "
                "BookCorpus, CommonCrawl, Pile, and PushShift.io corpora."
            ),
            "params": 2700000000,
            "path": "opt",
        },
        "kaggle_handle": "kaggle://keras/opt/keras/opt_2.7b_en/3",
    },
    "opt_6.7b_en": {
        "metadata": {
            "description": (
                "32-layer OPT model where case in maintained. Trained on "
                "BookCorpus, CommonCrawl, Pile, and PushShift.io corpora."
            ),
            "params": 6700000000,
            "path": "opt",
        },
        "kaggle_handle": "kaggle://keras/opt/keras/opt_6.7b_en/3",
    },
}



import numpy as np

from keras_hub.src.models.gemma.gemma_backbone import GemmaBackbone
from keras_hub.src.utils.preset_utils import get_file

backbone_cls = GemmaBackbone


def convert_backbone_config(transformers_config):
    backbone_config = dict()
    if transformers_config["model_type"] == "gemma":
        # Build Gemma backbone configuration
        backbone_config = {
            "vocabulary_size": transformers_config["vocab_size"],
            "num_layers": transformers_config["num_hidden_layers"],
            "num_query_heads": transformers_config["num_attention_heads"],
            "num_key_value_heads": transformers_config["num_key_value_heads"],
            "hidden_dim": transformers_config["hidden_size"],
            "intermediate_dim": transformers_config["intermediate_size"] * 2,
            "head_dim": transformers_config["head_dim"],
        }
    elif transformers_config["model_type"] == "gemma2":
        # Build Gemma 2 backbone configuration
        backbone_config = {
            "vocabulary_size": transformers_config["vocab_size"],
            "num_layers": transformers_config["num_hidden_layers"],
            "num_query_heads": transformers_config["num_attention_heads"],
            "num_key_value_heads": transformers_config["num_key_value_heads"],
            "hidden_dim": transformers_config["hidden_size"],
            "intermediate_dim": transformers_config["intermediate_size"] * 2,
            "head_dim": transformers_config["head_dim"],
            "query_head_dim_normalize": (
                transformers_config["head_dim"]
                == transformers_config["query_pre_attn_scalar"]
            ),
            "use_post_ffw_norm": True,
            "use_post_attention_norm": True,
            "final_logit_soft_cap": transformers_config[
                "final_logit_softcapping"
            ],
            "attention_logit_soft_cap": transformers_config[
                "attn_logit_softcapping"
            ],
            "sliding_window_size": transformers_config["sliding_window"],
            "use_sliding_window_attention": True,
        }
    return backbone_config


def convert_weights(backbone, loader, transformers_config):
    # Embedding layer
    loader.port_weight(
        keras_variable=backbone.get_layer("token_embedding").embeddings,
        hf_weight_key="model.embed_tokens.weight",
    )

    # Attention blocks
    for i in range(backbone.num_layers):
        decoder_layer = backbone.get_layer(f"decoder_block_{i}")
        # Norm layers
        loader.port_weight(
            keras_variable=decoder_layer.pre_attention_norm.scale,
            hf_weight_key=f"model.layers.{i}.input_layernorm.weight",
        )

        if decoder_layer.use_post_attention_norm:
            loader.port_weight(
                keras_variable=decoder_layer.post_attention_norm.scale,
                hf_weight_key=f"model.layers.{i}.post_attention_layernorm.weight",
            )

        if transformers_config["model_type"] == "gemma":
            loader.port_weight(
                keras_variable=decoder_layer.pre_ffw_norm.scale,
                hf_weight_key=f"model.layers.{i}.post_attention_layernorm.weight",
            )
        elif transformers_config["model_type"] == "gemma2":
            loader.port_weight(
                keras_variable=decoder_layer.pre_ffw_norm.scale,
                hf_weight_key=f"model.layers.{i}.pre_feedforward_layernorm.weight",
            )

        if decoder_layer.use_post_ffw_norm:
            loader.port_weight(
                keras_variable=decoder_layer.post_ffw_norm.scale,
                hf_weight_key=f"model.layers.{i}.post_feedforward_layernorm.weight",
            )

        # Attention layers
        loader.port_weight(
            keras_variable=decoder_layer.attention.query_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.q_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[0], keras_shape[2], keras_shape[1]),
                ),
                axes=(0, 2, 1),
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer.attention.key_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.k_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[0], keras_shape[2], keras_shape[1]),
                ),
                axes=(0, 2, 1),
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer.attention.value_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.v_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[0], keras_shape[2], keras_shape[1]),
                ),
                axes=(0, 2, 1),
            ),
        )
        loader.port_weight(
            keras_variable=decoder_layer.attention.output_dense.kernel,
            hf_weight_key=f"model.layers.{i}.self_attn.o_proj.weight",
            hook_fn=lambda hf_tensor, keras_shape: np.transpose(
                np.reshape(
                    hf_tensor,
                    (keras_shape[2], keras_shape[0], keras_shape[1]),
                ),
                axes=(1, 2, 0),
            ),
        )

        # MLP layers
        loader.port_weight(
            keras_variable=decoder_layer.gating_ffw.kernel,
            hf_weight_key=f"model.layers.{i}.mlp.gate_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer.gating_ffw_2.kernel,
            hf_weight_key=f"model.layers.{i}.mlp.up_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )
        loader.port_weight(
            keras_variable=decoder_layer.ffw_linear.kernel,
            hf_weight_key=f"model.layers.{i}.mlp.down_proj.weight",
            hook_fn=lambda hf_tensor, _: np.transpose(hf_tensor, axes=(1, 0)),
        )

    # Final normalization layer
    loader.port_weight(
        keras_variable=backbone.get_layer("final_normalization").scale,
        hf_weight_key="model.norm.weight",
    )


def convert_tokenizer(cls, preset, **kwargs):
    return cls(get_file(preset, "tokenizer.model"), **kwargs)



"""XLM-RoBERTa model preset configurations."""

backbone_presets = {
    "xlm_roberta_base_multi": {
        "metadata": {
            "description": (
                "12-layer XLM-RoBERTa model where case is maintained. "
                "Trained on CommonCrawl in 100 languages."
            ),
            "params": 277450752,
            "path": "xlm_roberta",
        },
        "kaggle_handle": "kaggle://keras/xlm_roberta/keras/xlm_roberta_base_multi/3",
    },
    "xlm_roberta_large_multi": {
        "metadata": {
            "description": (
                "24-layer XLM-RoBERTa model where case is maintained. "
                "Trained on CommonCrawl in 100 languages."
            ),
            "params": 558837760,
            "path": "xlm_roberta",
        },
        "kaggle_handle": "kaggle://keras/xlm_roberta/keras/xlm_roberta_large_multi/3",
    },
}



import keras
from keras import ops


class TimestepEmbedding(keras.layers.Layer):
    """Creates sinusoidal timestep embeddings.

    Call arguments:
        t: Tensor of shape (N,), representing N indices, one per batch element.
            These values may be fractional.
        dim: int. The dimension of the output.
        max_period: int, optional. Controls the minimum frequency of the
            embeddings. Defaults to 10000.
        time_factor: float, optional. A scaling factor applied to `t`. Defaults
            to 1000.0.

    Returns:
        A tensor of shape (N, D) representing the positional embeddings,
        where N is the number of batch elements and D is the specified
        dimension `dim`.
    """

    def call(self, t, dim, max_period=10000, time_factor=1000.0):
        t = time_factor * t
        half_dim = dim // 2
        freqs = ops.exp(
            ops.cast(-ops.log(max_period), dtype=t.dtype)
            * ops.arange(half_dim, dtype=t.dtype)
            / half_dim
        )
        args = t[:, None] * freqs[None]
        embedding = ops.concatenate([ops.cos(args), ops.sin(args)], axis=-1)

        if dim % 2 != 0:
            embedding = ops.concatenate(
                [embedding, ops.zeros_like(embedding[:, :1])], axis=-1
            )

        return embedding


class RotaryPositionalEmbedding(keras.layers.Layer):
    """
    Applies Rotary Positional Embedding (RoPE) to the input tensor.

    Call arguments:
        pos: KerasTensor. The positional tensor with shape (..., n, d).
        dim: int. The embedding dimension, should be even.
        theta: int. The base frequency.

    Returns:
        KerasTensor: The tensor with applied RoPE transformation.
    """

    def call(self, pos, dim, theta):
        scale = ops.arange(0, dim, 2, dtype="float32") / dim
        omega = 1.0 / (theta**scale)
        out = ops.einsum("...n,d->...nd", pos, omega)
        out = ops.stack(
            [ops.cos(out), -ops.sin(out), ops.sin(out), ops.cos(out)], axis=-1
        )
        out = ops.reshape(out, ops.shape(out)[:-1] + (2, 2))
        return ops.cast(out, dtype="float32")


class ApplyRoPE(keras.layers.Layer):
    """
    Applies the RoPE transformation to the query and key tensors.

    Call arguments:
        xq: KerasTensor. The query tensor of shape (..., L, D).
        xk: KerasTensor. The key tensor of shape (..., L, D).
        freqs_cis: KerasTensor. The frequency complex numbers tensor with shape
            `(..., 2)`.

    Returns:
        tuple[KerasTensor, KerasTensor]: The transformed query and key tensors.
    """

    def call(self, xq, xk, freqs_cis):
        xq_ = ops.reshape(xq, (*ops.shape(xq)[:-1], -1, 1, 2))
        xk_ = ops.reshape(xk, (*ops.shape(xk)[:-1], -1, 1, 2))

        xq_out = (
            freqs_cis[..., 0] * xq_[..., 0] + freqs_cis[..., 1] * xq_[..., 1]
        )
        xk_out = (
            freqs_cis[..., 0] * xk_[..., 0] + freqs_cis[..., 1] * xk_[..., 1]
        )

        return ops.reshape(xq_out, ops.shape(xq)), ops.reshape(
            xk_out, ops.shape(xk)
        )


class FluxRoPEAttention(keras.layers.Layer):
    """Computes the attention mechanism with RoPE.

    Args:
        dropout_p: float, optional. Dropout probability. Defaults to 0.0.
        is_causal: bool, optional. If True, applies causal masking. Defaults to
            False.

    Call arguments:
        q: KerasTensor. Query tensor of shape (..., L, D).
        k: KerasTensor. Key tensor of shape (..., S, D).
        v: KerasTensor. Value tensor of shape (..., S, D).
        positional_encoding: KerasTensor. Positional encoding tensor.

    Returns:
        KerasTensor: The resulting tensor from the attention mechanism.
    """

    def __init__(self, dropout_p=0.0, is_causal=False):
        super(FluxRoPEAttention, self).__init__()
        self.dropout_p = dropout_p
        self.is_causal = is_causal

    def call(self, q, k, v, positional_encoding):
        # Apply the RoPE transformation
        q, k = ApplyRoPE()(q, k, positional_encoding)

        # Scaled dot-product attention
        x = scaled_dot_product_attention(
            q, k, v, dropout_p=self.dropout_p, is_causal=self.is_causal
        )
        x = ops.transpose(x, (0, 2, 1, 3))
        b, s, h, d = ops.shape(x)
        return ops.reshape(x, (b, s, h * d))


# TODO: This is probably already implemented in several places, but is needed to
# ensure numeric equivalence to the original implementation. It uses
# torch.functional.scaled_dot_product_attention() - do we have an equivalent
# already in Keras?
def scaled_dot_product_attention(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    scale=None,
):
    """
    Computes the scaled dot-product attention.

    Args:
        query: KerasTensor. Query tensor of shape (..., L, D).
        key: KerasTensor. Key tensor of shape (..., S, D).
        value: KerasTensor. Value tensor of shape (..., S, D).
        attn_mask: KerasTensor, optional. Attention mask tensor. Defaults to
            None.
        dropout_p: float, optional. Dropout probability. Defaults to 0.0.
        is_causal: bool, optional. If True, applies causal masking. Defaults to
            False.
        scale: float, optional. Scale factor for attention. Defaults to None.

    Returns:
        KerasTensor: The output tensor from the attention mechanism.
    """
    L, S = ops.shape(query)[-2], ops.shape(key)[-2]
    scale_factor = (
        1 / ops.sqrt(ops.cast(ops.shape(query)[-1], dtype=query.dtype))
        if scale is None
        else scale
    )
    attn_bias = ops.zeros((L, S), dtype=query.dtype)

    if is_causal:
        assert attn_mask is None
        temp_mask = ops.ones((L, S), dtype=ops.bool)
        temp_mask = ops.tril(temp_mask, diagonal=0)
        attn_bias = ops.where(temp_mask, attn_bias, float("-inf"))

    if attn_mask is not None:
        if ops.shape(attn_mask)[-1] == 1:  # If the mask is 3D
            attn_bias += attn_mask
        else:
            attn_bias = ops.where(attn_mask, attn_bias, float("-inf"))

    # Compute attention weights
    attn_weight = (
        ops.matmul(query, ops.transpose(key, axes=[0, 1, 3, 2])) * scale_factor
    )
    attn_weight += attn_bias
    attn_weight = keras.activations.softmax(attn_weight, axis=-1)

    if dropout_p > 0.0:
        attn_weight = keras.layers.Dropout(dropout_p)(
            attn_weight, training=True
        )

    return ops.matmul(attn_weight, value)


def rearrange_symbolic_tensors(qkv, K, H):
    """
    Splits the qkv tensor into query (q), key (k), and value (v) components.

    Mimics rearrange(qkv, "B L (K H D) -> K B H L D", K=3, H=num_heads),
    for graph-mode TensorFlow support when doing functional subclassing
    models.

    Arguments:
        qkv: np.ndarray. Input tensor of shape (B, L, K*H*D).
        K: int. Number of components (q, k, v).
        H: int. Number of attention heads.

    Returns:
        tuple: q, k, v tensors of shape (B, H, L, D).
    """
    # Get the shape of qkv and calculate L and D
    B, L, dim = ops.shape(qkv)
    D = dim // (K * H)

    # Reshape and transpose the qkv tensor
    qkv_reshaped = ops.reshape(qkv, (B, L, K, H, D))
    qkv_transposed = ops.transpose(qkv_reshaped, (2, 0, 3, 1, 4))

    # Split q, k, v along the first dimension (K)
    qkv_splits = ops.split(qkv_transposed, K, axis=0)
    q, k, v = [ops.squeeze(split, 0) for split in qkv_splits]

    return q, k, v



import math

from keras import initializers
from keras import layers
from keras import ops

from keras_hub.src.layers.modeling.reversible_embedding import (
    ReversibleEmbedding,
)
from keras_hub.src.utils.keras_utils import clone_initializer
from keras_hub.src.utils.keras_utils import gelu_approximate
from keras_hub.src.utils.keras_utils import standardize_data_format


class SigLIPVisionEmbedding(layers.Layer):
    """A layer that converts images into patches.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        patch_size: int. The size of one side of each patch.
        image_size: int. The size of the input images.
        data_format: `None` or str. If specified, either `"channels_last"` or
            `"channels_first"`. The ordering of the dimensions in the
            inputs. `"channels_last"` corresponds to inputs with shape
            `(batch_size, height, width, channels)`
            while `"channels_first"` corresponds to inputs with shape
            `(batch_size, channels, height, width)`. It defaults to the
            `image_data_format` value found in your Keras config file at
            `~/.keras/keras.json`. If you never set it, then it will be
            `"channels_last"`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        hidden_dim,
        patch_size,
        image_size,
        data_format=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.patch_size = int(patch_size)
        self.image_size = int(image_size)
        self.data_format = standardize_data_format(data_format)
        self.num_positions = (image_size // patch_size) ** 2

        self.patch_embedding = layers.Conv2D(
            hidden_dim,
            kernel_size=patch_size,
            strides=patch_size,
            kernel_initializer=initializers.LecunNormal(),
            data_format=data_format,
            dtype=self.dtype_policy,
            name="patch_embedding",
        )
        self.position_embedding = layers.Embedding(
            self.num_positions,
            hidden_dim,
            embeddings_initializer=initializers.RandomNormal(
                stddev=1.0 / math.sqrt(hidden_dim)
            ),
            dtype=self.dtype_policy,
            name="position_embedding",
        )

    def build(self, input_shape):
        self.position_ids = self.add_weight(
            shape=(1, self.num_positions),
            initializer="zeros",
            # Let the backend determine the int dtype. For example, tf
            # requires int64 for correct device placement, whereas jax and torch
            # don't.
            dtype=int,
            trainable=False,
            name="position_ids",
        )
        self.position_ids.assign(
            ops.expand_dims(ops.arange(0, self.num_positions), axis=0)
        )
        self.patch_embedding.build(input_shape)
        self.position_embedding.build(self.position_ids.shape)

    def call(self, inputs, training=None):
        x = inputs
        batch_size = ops.shape(x)[0]
        patch_embeddings = self.patch_embedding(x, training=training)
        if self.data_format == "channels_last":
            patch_embeddings = ops.reshape(
                patch_embeddings, (batch_size, -1, self.hidden_dim)
            )
        else:
            patch_embeddings = ops.reshape(
                patch_embeddings, (batch_size, self.hidden_dim, -1)
            )
            patch_embeddings = ops.transpose(patch_embeddings, (0, 2, 1))
        position_embeddings = self.position_embedding(self.position_ids)
        return ops.add(patch_embeddings, position_embeddings)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "patch_size": self.patch_size,
                "image_size": self.image_size,
            }
        )
        return config

    def compute_output_shape(self, input_shape):
        output_shape = [input_shape[0], None, self.hidden_dim]
        if self.data_format == "channels_last":
            if input_shape[1] is not None and input_shape[2] is not None:
                patch_num = input_shape[1] // self.patch_size
                output_shape[1] = patch_num**2 + 1
        else:
            if input_shape[2] is not None and input_shape[3] is not None:
                patch_num = input_shape[2] // self.patch_size
                output_shape[1] = patch_num**2 + 1
        return output_shape


class SigLIPTextEmbedding(layers.Layer):
    """A layer which sums a token and position embedding.

    Args:
        vocabulary_size: The size of the vocabulary.
        sequence_length: The maximum length of input sequence.
        embedding_dim: The output dimension of the embedding layer
        tie_weights: Boolean, whether or not the matrix for embedding and
            the matrix for the `reverse` projection should share the same
            weights. Defaults to `True`.
        embeddings_initializer: The initializer to use for the Embedding
            Layers. Defaults to `"normal"`.
        mask_zero: Boolean, whether or not the input value 0 is a special
            "padding" value that should be masked out.
            This is useful when using recurrent layers which may take variable
            length input. If this is True, then all subsequent layers in the
            model need to support masking or an exception will be raised.
            If mask_zero` is set to True, as a consequence, index 0 cannot be
            used in the vocabulary
            (input_dim should equal size of vocabulary + 1). Defaults to
            `False`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `trainable`, `dtype` etc.
    """

    def __init__(
        self,
        vocabulary_size,
        sequence_length,
        embedding_dim,
        tie_weights=True,
        embeddings_initializer="normal",
        mask_zero=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.vocabulary_size = int(vocabulary_size)
        self.sequence_length = int(sequence_length)
        self.embedding_dim = int(embedding_dim)
        self.embeddings_initializer = initializers.get(embeddings_initializer)
        self.token_embedding = ReversibleEmbedding(
            vocabulary_size,
            embedding_dim,
            tie_weights=tie_weights,
            embeddings_initializer=clone_initializer(
                self.embeddings_initializer
            ),
            mask_zero=mask_zero,
            dtype=self.dtype_policy,
            name="token_embedding",
        )
        self.position_embedding = ReversibleEmbedding(
            sequence_length,
            embedding_dim,
            tie_weights=tie_weights,
            embeddings_initializer=clone_initializer(
                self.embeddings_initializer
            ),
            mask_zero=mask_zero,
            dtype=self.dtype_policy,
            name="position_embedding",
        )
        self.supports_masking = self.token_embedding.supports_masking

    def build(self, input_shape):
        input_shape = tuple(input_shape)
        self.token_embedding.build(input_shape)
        self.position_embedding.build((1, self.sequence_length))
        self.position_ids = self.add_weight(
            shape=(1, self.sequence_length),
            initializer="zeros",
            # Let the backend determine the int dtype. For example, tf
            # requires int64 for correct device placement, whereas jax and torch
            # don't.
            dtype=int,
            trainable=False,
            name="position_ids",
        )
        self.position_ids.assign(
            ops.expand_dims(ops.arange(0, self.sequence_length), axis=0)
        )

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "vocabulary_size": self.vocabulary_size,
                "sequence_length": self.sequence_length,
                "embedding_dim": self.embedding_dim,
                "embeddings_initializer": initializers.serialize(
                    self.embeddings_initializer
                ),
                "tie_weights": self.token_embedding.tie_weights,
                "mask_zero": self.token_embedding.mask_zero,
            }
        )
        return config

    def call(self, inputs):
        embedded_tokens = self.token_embedding(inputs)
        embedded_positions = self.position_embedding(self.position_ids)
        outputs = ops.add(embedded_tokens, embedded_positions)
        return outputs

    def compute_mask(self, inputs, mask=None):
        return self.token_embedding.compute_mask(inputs, mask=mask)

    def compute_output_shape(self, input_shape):
        return tuple(input_shape) + (self.embedding_dim,)


class SigLIPMLP(layers.Layer):
    """A SigLIP MLP block.

    Args:
        hidden_dim: int. The number of units in the output layer.
        intermediate_dim: int. The number of units in the intermediate layer.
        activation: str of callable. Activation to use in the intermediate
            layer.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(self, hidden_dim, intermediate_dim, activation, **kwargs):
        super().__init__(**kwargs)
        self.hidden_dim = hidden_dim
        self.intermediate_dim = intermediate_dim
        self.activation = activation

        if activation == "gelu_approximate":
            activation = gelu_approximate

        self.fc1 = layers.Dense(
            self.intermediate_dim,
            activation=activation,
            bias_initializer=initializers.RandomNormal(stddev=1e-6),
            dtype=self.dtype_policy,
            name="fc1",
        )
        self.fc2 = layers.Dense(
            self.hidden_dim,
            bias_initializer=initializers.RandomNormal(stddev=1e-6),
            dtype=self.dtype_policy,
            name="fc2",
        )

    def build(self, inputs_shape):
        self.fc1.build(inputs_shape)
        inputs_shape = self.fc1.compute_output_shape(inputs_shape)
        self.fc2.build(inputs_shape)

    def call(self, inputs):
        hidden_states = self.fc1(inputs)
        return self.fc2(hidden_states)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "activation": self.activation,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        outputs_shape = list(inputs_shape)
        outputs_shape[-1] = self.hidden_dim
        return outputs_shape


class SigLIPEncoderLayer(layers.Layer):
    """A SigLIP encoder layer.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        num_heads: int. Number of attention heads.
        intermediate_dim: int. The number of units in the intermediate layers.
        intermediate_activation: str or callable. Activation to use in the
            hidden layers.
        layer_norm_epsilon: float. The epsilon for the layer normalization.
            Defaults to `1e-6`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        hidden_dim,
        num_heads,
        intermediate_dim,
        intermediate_activation="gelu_approximate",
        use_causal_mask=False,
        layer_norm_epsilon=1e-6,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if hidden_dim % num_heads != 0:
            raise ValueError(
                "`hidden_dim` must be divisible by `num_heads`. "
                f"Received: hidden_dim={hidden_dim}, num_heads={num_heads}"
            )
        self.hidden_dim = int(hidden_dim)
        self.num_heads = int(num_heads)
        self.intermediate_dim = int(intermediate_dim)
        self.intermediate_activation = intermediate_activation
        self.use_causal_mask = bool(use_causal_mask)
        self.layer_norm_epsilon = layer_norm_epsilon

        self.self_attn = layers.MultiHeadAttention(
            num_heads,
            hidden_dim // num_heads,
            dtype=self.dtype_policy,
            name="self_attn",
        )
        self.layer_norm1 = layers.LayerNormalization(
            epsilon=layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="layer_norm1",
        )
        self.mlp = SigLIPMLP(
            hidden_dim,
            intermediate_dim,
            intermediate_activation,
            dtype=self.dtype_policy,
            name="mlp",
        )
        self.layer_norm2 = layers.LayerNormalization(
            epsilon=layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="layer_norm2",
        )

    def build(self, inputs_shape):
        self.layer_norm1.build(inputs_shape)
        self.self_attn.build(inputs_shape, inputs_shape, inputs_shape)
        self.layer_norm2.build(inputs_shape)
        self.mlp.build(inputs_shape)

    def compute_output_shape(self, inputs_shape):
        outputs_shape = list(inputs_shape)
        outputs_shape[-1] = self.hidden_dim
        return outputs_shape

    def call(self, inputs, training=None):
        residual = inputs
        x = self.layer_norm1(inputs)
        x = self.self_attn(
            x, x, x, training=training, use_causal_mask=self.use_causal_mask
        )
        x = ops.add(residual, x)

        residual = x
        x = self.layer_norm2(x)
        x = self.mlp(x, training=training)
        x = ops.add(residual, x)
        return x

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "num_heads": self.num_heads,
                "intermediate_dim": self.intermediate_dim,
                "intermediate_activation": self.intermediate_activation,
                "use_causal_mask": self.use_causal_mask,
                "layer_norm_epsilon": self.layer_norm_epsilon,
            }
        )
        return config


class SigLIPMultiHeadAttentionPooling(layers.Layer):
    """A SigLIP multi-headed attention pooling layer.

    Args:
        hidden_dim: int. The number of units in the hidden layers.
        intermediate_dim: int. The number of units in the intermediate layers.
        num_heads: int. Number of attention heads.
        activation: str or callable. Activation to use in the MLP.
        layer_norm_epsilon: float. The epsilon for the layer normalization.
            Defaults to `1e-6`.
        **kwargs: other keyword arguments passed to `keras.layers.Layer`,
            including `name`, `dtype` etc.
    """

    def __init__(
        self,
        hidden_dim,
        intermediate_dim,
        num_heads,
        activation,
        layer_norm_epsilon=1e-6,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.hidden_dim = int(hidden_dim)
        self.intermediate_dim = int(intermediate_dim)
        self.num_heads = int(num_heads)
        self.activation = activation
        self.layer_norm_epsilon = layer_norm_epsilon

        self.attention = layers.MultiHeadAttention(
            num_heads,
            hidden_dim // num_heads,
            dtype=self.dtype_policy,
            name="attention",
        )
        self.layer_norm = layers.LayerNormalization(
            epsilon=self.layer_norm_epsilon,
            dtype=self.dtype_policy,
            name="layernorm",
        )
        self.mlp = SigLIPMLP(
            hidden_dim=self.hidden_dim,
            intermediate_dim=self.intermediate_dim,
            activation=self.activation,
            dtype=self.dtype_policy,
            name="mlp",
        )

    def build(self, inputs_shape):
        self.probe = self.add_weight(
            (1, 1, self.hidden_dim),
            initializer=initializers.GlorotUniform(),
            dtype=self.dtype_policy.variable_dtype,
        )
        self.attention.build(
            query_shape=(inputs_shape[0], 1, self.hidden_dim),
            value_shape=inputs_shape,
        )
        inputs_shape = self.attention.compute_output_shape(
            query_shape=(inputs_shape[0], 1, self.hidden_dim),
            value_shape=inputs_shape,
        )
        self.layer_norm.build(inputs_shape)
        self.mlp.build(inputs_shape)

    def call(self, inputs, training=None):
        batch_size = ops.shape(inputs)[0]
        probes = ops.repeat(self.probe, repeats=batch_size, axis=0)
        hidden_states = self.attention(
            probes, inputs, inputs, training=training
        )
        residuals = hidden_states
        hidden_states = self.layer_norm(hidden_states)
        hidden_states = ops.add(residuals, self.mlp(hidden_states))
        return hidden_states[:, 0]

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "hidden_dim": self.hidden_dim,
                "intermediate_dim": self.intermediate_dim,
                "num_heads": self.num_heads,
                "activation": self.activation,
                "layer_norm_epsilon": self.layer_norm_epsilon,
            }
        )
        return config

    def compute_output_shape(self, inputs_shape):
        return (inputs_shape[0], self.hidden_dim)


class SigLIPHead(layers.Layer):
    """The head layer of SigLIP.

    `SigLIP` takes `vision_embedding` and `text_embedding` as inputs to
    compute the corresponding logits. Both embeddings are L2 normalized and used
    to compute pairwise cosine similarity. The resulting logits are then scaled
    and added by learnable `logit_scale`  and `logit_bias` parameters.

    Call arguments:
        vision_embedding: A tensor of shape `(batch_size, hidden_dim)`.
        text_embedding: A tensor of shape `(batch_size, hidden_dim)`.
    """

    def build(self, input_shape):
        self.logit_scale = self.add_weight(
            shape=(),
            initializer=initializers.Constant(math.log(1.0)),
            trainable=True,
            dtype=self.variable_dtype,
            name="logit_scale",
        )
        self.logit_bias = self.add_weight(
            shape=(),
            initializer=initializers.Zeros(),
            trainable=True,
            dtype=self.variable_dtype,
            name="logit_bias",
        )

    def call(self, vision_embedding, text_embedding):
        normalized_vision_embedding = ops.sqrt(
            ops.sum(ops.power(vision_embedding, 2), axis=-1, keepdims=True)
        )
        normalized_text_embedding = ops.sqrt(
            ops.sum(ops.power(text_embedding, 2), axis=-1, keepdims=True)
        )
        vision_embedding = ops.divide(
            vision_embedding, normalized_vision_embedding
        )
        text_embedding = ops.divide(text_embedding, normalized_text_embedding)
        text_logits = ops.add(
            ops.multiply(
                ops.matmul(text_embedding, ops.transpose(vision_embedding)),
                ops.exp(self.logit_scale),
            ),
            self.logit_bias,
        )
        vision_logits = ops.transpose(text_logits)
        return vision_logits, text_logits

    def compute_output_shape(
        self, vision_embedding_shape, text_embedding_shape
    ):
        vision_logits_shape = (
            vision_embedding_shape[0],
            text_embedding_shape[0],
        )
        text_logits_shape = (
            text_embedding_shape[0],
            vision_embedding_shape[0],
        )
        return vision_logits_shape, text_logits_shape



import contextlib
import functools
import inspect
import math
import re
import threading

import keras
import numpy as np
from keras import ops
from packaging import version

try:
    import tensorflow as tf
    import tensorflow_text as tf_text
except ImportError:
    tf = None
    tf_text = None


NO_CONVERT_COUNTER = threading.local()


@contextlib.contextmanager
def no_convert_scope():
    try:
        NO_CONVERT_COUNTER.count = getattr(NO_CONVERT_COUNTER, "count", 0) + 1
        yield
    finally:
        NO_CONVERT_COUNTER.count = getattr(NO_CONVERT_COUNTER, "count", 0) - 1


def in_tf_function():
    if tf is None:
        return False
    return not tf.executing_eagerly()


def in_no_convert_scope():
    return getattr(NO_CONVERT_COUNTER, "count", 0) > 0


def preprocessing_function(fn):
    """Wraps a preprocessing function to handle tf tensor conversion."""
    if tf is None:
        return fn

    params = inspect.signature(fn).parameters
    accepts_labels = all(k in params for k in ("x", "y", "sample_weight"))
    if not accepts_labels:

        @functools.wraps(fn)
        def wrapper(self, x, **kwargs):
            with tf.device("cpu"):
                x = convert_preprocessing_inputs(x)
                with no_convert_scope():
                    x = fn(self, x, **kwargs)
                return convert_preprocessing_outputs(x)

    else:

        @functools.wraps(fn)
        def wrapper(self, x, y=None, sample_weight=None, **kwargs):
            with tf.device("cpu"):
                x, y, sample_weight = convert_preprocessing_inputs(
                    (x, y, sample_weight)
                )
                with no_convert_scope():
                    x = fn(self, x, y=y, sample_weight=sample_weight, **kwargs)
                return convert_preprocessing_outputs(x)

    return wrapper


def convert_preprocessing_inputs(x):
    """Convert raw inputs for preprocessing.

    This function is used to convert raw inputs (strings, lists, `np.ndarray`s,
    `jax.Array`s, `torch.Tensor`s, etc) to a canonical format for
    preprocessing layers. All inputs will be converted to backend tensors if
    possible, except ragged inputs and string inputs which be converted to tf
    tensors regardless of backend.

    `tuple` and `list` elements are handled differently by this function. A
    `tuple` is assumed to enumerate separate inputs, and a `list` is assumed to
    enumerate elements in a single array-like input. This makes it possible to
    represent ragged and string inputs in a multi-backend format, as shown in
    the examples below.

    Examples:
    ```python
    # Two ragged arrays of token ids.
    x = ([[1, 2, 3], [4, 5]], [[1, 2], [3, 4, 5]])
    keras_hub.utils.convert_preprocessing_inputs(x)

    # A batch of three samples each with two string segments.
    x = (["hi", "hello", "hey"], ["bye", "later", "so long"])
    keras_hub.utils.convert_preprocessing_inputs(x)

    # A batch of features in a dictionary.
    x = {
        "text": ["hi", "hello", "hey"],
        "images": np.ones((3, 64, 64, 3)),
        "labels": [1, 0, 1],
    }
    keras_hub.utils.convert_preprocessing_inputs(x)
    ```
    """
    if not tf.executing_eagerly() or in_no_convert_scope():
        return x

    if isinstance(x, dict):
        return {k: convert_preprocessing_inputs(x[k]) for k, v in x.items()}
    if isinstance(x, tuple):
        return tuple(convert_preprocessing_inputs(v) for v in x)
    if isinstance(x, (str, bytes)):
        return tf.constant(x)
    if isinstance(x, list):
        try:
            numpy_x = np.array(x)
        except ValueError as e:
            # If numpy conversion failed, try converting to a ragged array.
            try:
                return tf.ragged.constant(x)
            except ValueError:
                # If ragged conversion failed return to the numpy error.
                raise e
        # If we have a string input, use tf.tensor.
        if numpy_x.dtype.type is np.str_ or numpy_x.dtype.type is np.bytes_:
            return tf.convert_to_tensor(x)
        # Numpy will default to int64, int32 works with more ops.
        if numpy_x.dtype == np.int64:
            numpy_x = numpy_x.astype(np.int32)
        # We have non-ragged, non-string input. Use backbend type.
        x = ops.convert_to_tensor(numpy_x)
        # Torch will complain about device placement for GPU tensors.
        if keras.config.backend() == "torch":
            x = x.cpu()
        return x
    if is_tensor_type(x):
        # String or ragged types we keep as tf.
        if isinstance(x, tf.RaggedTensor) or x.dtype == tf.string:
            return x
        # If we have a string input, use tf.tensor.
        if isinstance(x, np.ndarray) and x.dtype.type is np.str_:
            return tf.convert_to_tensor(x)
        x = ops.convert_to_tensor(x)
        # Torch will complain about device placement for GPU tensors.
        if keras.config.backend() == "torch":
            x = x.cpu()
        return x
    return x


def convert_preprocessing_outputs(x):
    """Convert outputs after preprocessing to a backend agnostic format.

    This function is used to convert `tf.Tensor` and `tf.RaggedTensor` output
    from preprocessing layers to either:

    - The correct tensor type for the Keras backend framework.
    - Python lists, in the case of ragged and string data.

    This will automatically be called when on the output of preprocessing
    layers or `keras_hub.models.Task`s with preprocessing included. It could be
    used directly to convert a `tf.data.Dataset` output to a backend agnostic
    type.

    Examples:
    ```python
    # Two ragged arrays of token ids.
    x = tf.ragged.constant([[1, 2, 3], [4, 5]])
    keras_hub.utils.convert_preprocessing_outputs(x)

    # A batch of three samples each with two string segments.
    x = (tf.constant["hi", "yo", "hey"]), tf.constant(["bye", "ciao", ""]))
    keras_hub.utils.convert_preprocessing_outputs(x)

    # A batch of features in a dictionary.
    x = {
        "text": tf.constant(["hi", "hello", "hey"]),
        "images": tf.ones((3, 64, 64, 3)),
        "labels": tf.constant([1, 0, 1]),
    }
    keras_hub.utils.convert_preprocessing_outputs(x)
    ```
    """
    if not tf.executing_eagerly() or in_no_convert_scope():
        return x

    def convert(x):
        if x is None:
            return x
        if isinstance(x, tf.RaggedTensor) or x.dtype == tf.string:
            return tensor_to_list(x)
        dtype = keras.backend.standardize_dtype(x.dtype)
        return ops.convert_to_tensor(x, dtype=dtype)

    return keras.tree.map_structure(convert, x)


def _decode_strings_to_utf8(inputs):
    """Recursively decodes to list of strings with 'utf-8' encoding."""
    if isinstance(inputs, bytes):
        # Handles the case when the input is a scalar string.
        return inputs.decode("utf-8", errors="ignore")
    else:
        # Recursively iterate when input is a list.
        return [_decode_strings_to_utf8(x) for x in inputs]


def tensor_to_list(inputs):
    """Converts a tensor to nested lists.

    Args:
        inputs: Input tensor, or dict/list/tuple of input tensors.
    """
    if not isinstance(inputs, (tf.RaggedTensor, tf.Tensor)):
        inputs = tf.convert_to_tensor(inputs)
    if isinstance(inputs, tf.RaggedTensor):
        list_outputs = inputs.to_list()
    elif isinstance(inputs, tf.Tensor):
        list_outputs = inputs.numpy()
        if inputs.shape.rank != 0:
            list_outputs = list_outputs.tolist()
    if inputs.dtype == tf.string:
        list_outputs = _decode_strings_to_utf8(list_outputs)
    return list_outputs


def convert_to_ragged_batch(inputs):
    """Ensure a tf.Tensor is a ragged rank 2 tensor."""
    if not isinstance(inputs, (tf.RaggedTensor, tf.Tensor)):
        inputs = tf.convert_to_tensor(inputs)
    unbatched = inputs.shape.rank == 1
    rectangular = isinstance(inputs, tf.Tensor)
    if unbatched:
        inputs = tf.expand_dims(inputs, 0)
    if rectangular:
        inputs = tf.RaggedTensor.from_tensor(inputs)
    return inputs, unbatched, rectangular


def truncate_at_token(inputs, token, mask):
    """Truncate at first instance of `token`, ignoring `mask`."""
    matches = (inputs == token) & (~mask)
    end_indices = tf.cast(tf.math.argmax(matches, -1), "int32")
    end_indices = tf.where(end_indices == 0, tf.shape(inputs)[-1], end_indices)
    return tf.RaggedTensor.from_tensor(inputs, end_indices)


def strip_to_ragged(token_ids, mask, ids_to_strip):
    """Remove masked and special tokens from a sequence before detokenizing."""
    mask = tf.cast(mask, "bool")
    for id in ids_to_strip:
        mask = mask & (token_ids != id)
    return tf.ragged.boolean_mask(token_ids, mask)


def assert_tf_libs_installed(symbol_name):
    if tf_text is None or tf is None:
        raise ImportError(
            f"{symbol_name} requires `tensorflow` and `tensorflow-text` for "
            "text processing. Run `pip install tensorflow-text` to install "
            "both packages or visit https://www.tensorflow.org/install\n\n"
            "If `tensorflow-text` is already installed, try importing it "
            "in a clean python session. Your installation may have errors.\n\n"
            "KerasHub uses `tf.data` and `tensorflow-text` to preprocess text "
            "on all Keras backends. If you are running on Jax or Torch, this "
            "installation does not need GPU support."
        )


def check_bounding_box_support():
    return version.parse(keras.__version__) >= version.parse("3.8.0")


def assert_bounding_box_support(symbol_name):
    if not check_bounding_box_support():
        raise ImportError(
            f"{symbol_name} requires Keras version to be 3.8.0 or higher. "
            f"Current keras version: {keras.__version__}"
        )


def assert_tf_backend(symbol_name):
    if keras.config.backend() != "tensorflow":
        raise RuntimeError(
            f"{symbol_name} requires the `tensorflow` backend. "
            "Please set `KERAS_BACKEND=tensorflow` when running your program."
        )


def is_tensor_type(x):
    return hasattr(x, "__array__")


def is_float_dtype(dtype):
    return "float" in keras.backend.standardize_dtype(dtype)


def is_int_dtype(dtype):
    return "int" in keras.backend.standardize_dtype(dtype)


def is_string_dtype(dtype):
    return "string" in keras.backend.standardize_dtype(dtype)


def get_dtype_size_in_bits(dtype):
    """Get the size of a given dtype in bits."""
    dtype = keras.backend.standardize_dtype(dtype)
    # If dtype is bool, return 1 immediately.
    if dtype == "bool":
        return 1
    # Else, we extract the bit size from the string.
    return int(re.sub(r"bfloat|float|uint|int", "", dtype))


def get_tensor_size_in_bits(shape, dtype):
    """Calculate the size given dtype and shape in bits.

    Args:
        dtype: The dtype of the tensor.
        shape: List of iterables representing the shape of the tensor.

    Returns:
        The size of the tensor in bytes.
    """
    return math.prod(shape) * get_dtype_size_in_bits(dtype)


def any_equal(inputs, values, padding_mask):
    """Return a mask that is True anywhere `inputs` has a value in `values`.

    Final mask has `padding_mask` applied.

    Args:
        inputs: Input tensor.
        values: List or iterable of tensors shaped like `inputs` or
            broadcastable by bit operators.
        padding_mask: Tensor with shape compatible with inputs that will
            condition output.

    Returns:
        A tensor with `inputs` shape where each position is True if it contains
            a value from any `values`. Padding mask will be applied before
            returning.
    """
    output = ops.equal(inputs, values[0])
    for value in values[1:]:
        value_equality = ops.equal(inputs, value)
        output = ops.logical_or(output, value_equality)

    return ops.logical_and(output, padding_mask)


def target_gather(
    targets,
    indices,
    mask=None,
    mask_val=0.0,
):
    """A utility function wrapping `ops.take`, which deals with:
        1) both batched and unbatched `targets`.
        2) when unbatched `targets` have empty rows, the result will be filled
            with `mask_val`.
        3) target masking.

    Args:
        targets: `[N, ...]` or `[batch_size, N, ...]` Tensor representing
            targets such as boxes, keypoints, etc.
        indices: [M] or [batch_size, M] int32 Tensor representing indices within
            `targets` to gather.
        mask: `[M, ...]` or `[batch_size, M, ...]` boolean Tensor
            representing the masking for each target. `True` means the
            corresponding entity should be masked to `mask_val`, `False`
            means the corresponding entity should be the target value.
            Defaults to `None`.
        mask_val: float. representing the masking value if `mask` is True
            on the entity.
            Defaults to `0.0`

    Returns:
        targets: `[M, ...]` or `[batch_size, M, ...]` Tensor representing
            selected targets.

        Raise:
            ValueError: If `targets` is higher than rank 3.
    """
    targets_shape = list(targets.shape)
    if len(targets_shape) > 3:
        raise ValueError(
            f"`target_gather` does not support `targets` with rank "
            f"larger than 3, got {len(targets.shape)}"
        )

    def gather_unbatched(labels, match_indices, mask, mask_val):
        """Gather based on unbatched labels and boxes."""
        num_gt_boxes = labels.shape[0]

        def assign_when_rows_empty():
            if len(labels.shape) > 1:
                mask_shape = [match_indices.shape[0], labels.shape[-1]]
            else:
                mask_shape = [match_indices.shape[0]]
            return ops.cast(mask_val, labels.dtype) * ops.ones(
                mask_shape, dtype=labels.dtype
            )

        def assign_when_rows_not_empty():
            targets = ops.take(labels, match_indices, axis=0)
            if mask is None:
                return targets
            else:
                masked_targets = ops.cast(
                    mask_val, labels.dtype
                ) * ops.ones_like(mask, dtype=labels.dtype)
                return ops.where(mask, masked_targets, targets)

        if num_gt_boxes > 0:
            return assign_when_rows_not_empty()
        else:
            return assign_when_rows_empty()

    def _gather_batched(labels, match_indices, mask, mask_val):
        """Gather based on batched labels."""
        batch_size = labels.shape[0]
        if batch_size == 1:
            if mask is not None:
                result = gather_unbatched(
                    ops.squeeze(labels, axis=0),
                    ops.squeeze(match_indices, axis=0),
                    ops.squeeze(mask, axis=0),
                    mask_val,
                )
            else:
                result = gather_unbatched(
                    ops.squeeze(labels, axis=0),
                    ops.squeeze(match_indices, axis=0),
                    None,
                    mask_val,
                )
            return ops.expand_dims(result, axis=0)
        else:
            targets = ops.take_along_axis(
                labels, ops.expand_dims(match_indices, axis=-1), axis=1
            )

            if mask is None:
                return targets
            else:
                masked_targets = ops.cast(
                    mask_val, labels.dtype
                ) * ops.ones_like(mask, dtype=labels.dtype)
                return ops.where(mask, masked_targets, targets)

    if len(targets_shape) <= 2:
        return gather_unbatched(targets, indices, mask, mask_val)
    elif len(targets_shape) == 3:
        return _gather_batched(targets, indices, mask, mask_val)



"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras_hub.src.utils.coco.coco_utils import (
    coco_id_to_name as coco_id_to_name,
)
from keras_hub.src.utils.coco.coco_utils import (
    coco_name_to_id as coco_name_to_id,
)
from keras_hub.src.utils.imagenet.imagenet_utils import (
    decode_imagenet_predictions as decode_imagenet_predictions,
)
from keras_hub.src.utils.imagenet.imagenet_utils import (
    imagenet_id_to_name as imagenet_id_to_name,
)
from keras_hub.src.utils.imagenet.imagenet_utils import (
    imagenet_name_to_id as imagenet_name_to_id,
)
